#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
#  import mlflow
import os
import re

import numpy as np
import torch
import torch.nn as nn
from loguru import logger
from sklearn.metrics import classification_report, roc_auc_score
from torch.nn import BCEWithLogitsLoss, CrossEntropyLoss, MSELoss
from tqdm import tqdm

from transformers import BertConfig, BertTokenizer, BertTokenizerFast
#  from transformers.modeling_bert import BertModel, BertPreTrainedModel
from transformers.models.bert import BertModel, BertPreTrainedModel

from ...losses import DiceLoss, FocalLoss
from ...utils import acc_and_f1, sigmoid, softmax
from ...utils.multiprocesses import (barrier_leader_process,
                                     barrier_member_processes)
from ..trainer import Trainer, get_default_optimizer_parameters


def logits_to_preds(logits):
    preds = np.argmax(logits, axis=1)
    probs = softmax(logits)
    # 调整类别0的判断阈值
    #  preds = np.array([
    #      0 if x == 1 and prob[1] < 0.60 else x for prob, x in zip(probs, preds)
    #  ])
    #  preds = np.array([
    #      1 if x == 0 and prob[1] < 0.90 else x for prob, x in zip(probs, preds)
    #  ])

    return preds, probs


class EmotionTokenizer(BertTokenizer):
    def __init__(self,
                 vocab_file,
                 do_lower_case=False,
                 is_english=False,
                 emotion_words_file=None):
        super().__init__(vocab_file=str(vocab_file),
                         do_lower_case=do_lower_case)
        self.vocab_file = str(vocab_file)
        self.do_lower_case = do_lower_case
        self.is_english = is_english

        self.emotion_words_file = emotion_words_file
        self.emotion_words = []
        self._init_emotion_words()
        self.emotion2id = {w: i for i, w in enumerate(self.emotion_words)}
        self.id2emotion = {i: w for i, w in enumerate(self.emotion_words)}

    def _init_emotion_words(self):
        logger.info(f"emotion_words_file: {self.emotion_words_file}")
        with open(self.emotion_words_file, 'r') as rd:
            lines = rd.readlines()
        emotion_words = set()
        for line in tqdm(lines, desc="Load emotion words"):
            line = line.strip()
            if line.startswith('#'):
                continue
            toks = line.split('\t')
            if len(toks) < 2:
                continue
            emotion_words.add(toks[0])
            if len(emotion_words) >= 100:
                break
        self.emotion_words = list(emotion_words)

    def _replace_emotions(self, tokens, max_seq_length):
        new_tokens = []
        for i in range(len(tokens)):
            found = False
            for j in range(len(self.emotion_words)):
                eword = self.emotion_words[j]
                tword = ''.join(tokens[i:i + len(eword)])
                if eword == tword:
                    new_tokens.append(f"[unused{j}]")
                    i += len(eword)
                    found = True
                    break
            if not found:
                new_tokens.append(tokens[i])
        return new_tokens

    def _append_emotions(self, tokens, max_seq_length):
        new_tokens = []
        for i in range(len(tokens)):
            found = False
            for j in range(len(self.emotion_words)):
                eword = self.emotion_words[j]
                tword = ''.join(tokens[i:i + len(eword)])
                if eword == tword:
                    new_tok = f"[unused{j}]"
                    if new_tok not in new_tokens:
                        new_tokens.append(new_tok)
                        i += len(eword)
                        found = True
                        break
        tail_tokens = []
        for tok in new_tokens:
            tail_tokens.append("[SEP]")
            tail_tokens.append(tok)

        if max_seq_length > 0:
            if len(tokens) + len(tail_tokens) > max_seq_length:
                tokens = tokens[:max_seq_length - len(tail_tokens)]

        tokens.extend(tail_tokens)
        #  logger.info(f"tokens: {tokens}")
        return tokens

    def _append_emotions_by_flash_text(self, tokens, max_seq_length):
        from flasttext import KeywordProcessor
        new_tokens = []
        for i in range(len(tokens)):
            found = False
            for j in range(len(self.emotion_words)):
                eword = self.emotion_words[j]
                tword = ''.join(tokens[i:i + len(eword)])
                if eword == tword:
                    new_tok = f"[unused{j}]"
                    if new_tok not in new_tokens:
                        new_tokens.append(new_tok)
                        i += len(eword)
                        found = True
                        break
        tail_tokens = []
        for tok in new_tokens:
            tail_tokens.append("[SEP]")
            tail_tokens.append(tok)

        if max_seq_length > 0:
            if len(tokens) + len(tail_tokens) > max_seq_length:
                tokens = tokens[:max_seq_length - len(tail_tokens)]

        tokens.extend(tail_tokens)
        return tokens

    def tokenize(self, text, max_seq_length=0):
        text_tokens = [c for c in text]

        _tokens = []
        for c in text_tokens:
            if self.do_lower_case:
                c = c.lower()

            if c in self.vocab:
                _tokens.append(c)
            else:
                _tokens.append('[UNK]')
        if self.emotion_words:
            #  _tokens = self._replace_emotions(_tokens, max_seq_length)
            _tokens = self._append_emotions(_tokens, max_seq_length)
        return _tokens


class BertForSequenceClassification(BertPreTrainedModel):
    def __init__(self, config):
        super().__init__(config)
        self.num_labels = config.num_labels

        self.bert = BertModel(config)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)
        self.classifier = nn.Linear(config.hidden_size, config.num_labels)
        self.loss_type = config.loss_type
        self.focalloss_gamma = config.focalloss_gamma
        self.focalloss_alpha = config.focalloss_alpha
        self.diceloss_weight = config.diceloss_weight

        self.init_weights()

    #  @add_start_docstrings_to_callable(BERT_INPUTS_DOCSTRING)
    def forward(
        self,
        input_ids=None,
        attention_mask=None,
        token_type_ids=None,
        position_ids=None,
        head_mask=None,
        inputs_embeds=None,
        labels=None,
    ):
        r"""
        labels (:obj:`torch.LongTensor` of shape :obj:`(batch_size,)`, `optional`, defaults to :obj:`None`):
            Labels for computing the sequence classification/regression loss.
            Indices should be in :obj:`[0, ..., config.num_labels - 1]`.
            If :obj:`config.num_labels == 1` a regression loss is computed (Mean-Square loss),
            If :obj:`config.num_labels > 1` a classification loss is computed (Cross-Entropy).

    Returns:
        :obj:`tuple(torch.FloatTensor)` comprising various elements depending on the configuration (:class:`~transformers.BertConfig`) and inputs:
        loss (:obj:`torch.FloatTensor` of shape :obj:`(1,)`, `optional`, returned when :obj:`label` is provided):
            Classification (or regression if config.num_labels==1) loss.
        logits (:obj:`torch.FloatTensor` of shape :obj:`(batch_size, config.num_labels)`):
            Classification (or regression if config.num_labels==1) scores (before SoftMax).
        hidden_states (:obj:`tuple(torch.FloatTensor)`, `optional`, returned when ``config.output_hidden_states=True``):
            Tuple of :obj:`torch.FloatTensor` (one for the output of the embeddings + one for the output of each layer)
            of shape :obj:`(batch_size, sequence_length, hidden_size)`.

            Hidden-states of the model at the output of each layer plus the initial embedding outputs.
        attentions (:obj:`tuple(torch.FloatTensor)`, `optional`, returned when ``config.output_attentions=True``):
            Tuple of :obj:`torch.FloatTensor` (one for each layer) of shape
            :obj:`(batch_size, num_heads, sequence_length, sequence_length)`.

            Attentions weights after the attention softmax, used to compute the weighted average in the self-attention
            heads.

    Examples::

        from transformers import BertTokenizer, BertForSequenceClassification
        import torch

        tokenizer = BertTokenizer.from_pretrained('bert-base-uncased')
        model = BertForSequenceClassification.from_pretrained('bert-base-uncased')

        input_ids = torch.tensor(tokenizer.encode("Hello, my dog is cute", add_special_tokens=True)).unsqueeze(0)  # Batch size 1
        labels = torch.tensor([1]).unsqueeze(0)  # Batch size 1
        outputs = model(input_ids, labels=labels)

        loss, logits = outputs[:2]

        """

        outputs = self.bert(
            input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
        )

        pooled_output = outputs[1]

        pooled_output = self.dropout(pooled_output)
        logits = self.classifier(pooled_output)

        outputs = (logits, ) + outputs[
            2:]  # add hidden states and attention if they are here

        #  if labels is not None:
        #  if labels is not None and any(labels):
        if labels is not None and torch.sum(labels) > 0:
            if self.num_labels == 1:
                #  We are doing regression
                loss_fct = MSELoss()
                loss = loss_fct(logits.view(-1), labels.view(-1))
            else:

                #  #  logger.debug(f"logits: {type(logits)}, {logits}")
                #  preds = logits_to_preds(logits.detach().cpu().numpy())
                #  preds = torch.tensor(preds).float().cuda()
                #  loss = FocalLoss(gamma=1.5, alpha=None)(preds.view(-1),
                #                                          labels.view(-1))
                #

                #  loss_fct = CrossEntropyLoss()
                #  loss = loss_fct(logits.view(-1, self.num_labels),
                #                  labels.view(-1))

                #  logger.debug(f"preds: {type(preds)}, {preds}")

                assert self.loss_type in [
                    'FocalLoss', 'DiceLoss', 'CircleLoss', 'CrossEntropyLoss',
                    'BCEWithLogitsLoss'
                ]
                if self.loss_type == 'FocalLoss':
                    # 0.58
                    loss = FocalLoss(gamma=self.focalloss_gamma,
                                     alpha=self.focalloss_alpha)(logits.view(
                                         -1, self.num_labels), labels.view(-1))
                elif self.loss_type == 'CircleLoss':
                    from ...losses.circleloss import CircleLoss
                    # 0.496885 (4/5)
                    loss = CircleLoss(gamma=15, m=1e-6, similarity='cos')(
                        logits.view(-1, self.num_labels), labels.view(-1))
                    #  loss = CircleLoss(gamma=15, m=1e-6, similarity='dot')(
                    #      logits.view(-1, self.num_labels), labels.view(-1))
                elif self.loss_type == 'DiceLoss':
                    #  loss = DiceLoss(weight=self.diceloss_weight)(logits.view(
                    #      -1, self.num_labels), labels.view(-1))
                    loss = DiceLoss(epsilon=1e-5)(logits.view(
                        -1, self.num_labels), labels.view(-1, self.num_labels))
                elif self.loss_type == 'BCEWithLogitsLoss':
                    loss_fct = BCEWithLogitsLoss()
                    #  logger.debug(f"logits: {logits.shape}, {logits}")
                    #  logger.debug(f"labels: {labels.shape}, {labels}")
                    #  loss = loss_fct(logits, labels.float())
                    loss = loss_fct(logits.view(-1, self.num_labels),
                                    labels.view(-1, self.num_labels).float())
                else:
                    loss_fct = CrossEntropyLoss()
                    #  logger.warning(f"logits: {logits.shape}")
                    #  logger.warning(f"labels: {labels.shape}")
                    loss = loss_fct(logits.view(-1, self.num_labels),
                                    labels.view(-1))
                    #  loss = loss_fct(logits.view(-1, self.num_labels),
                    #                  labels.view(-1, self.num_labels))

                #  from ...losses import FocalLoss
                #  #  loss = FocalLoss(gamma=1.5,
                #  #                   alpha=None)(logits.view(-1, self.num_labels),
                #  #                               labels.view(-1))
                #  loss = FocalLoss(
                #      gamma=1.5,
                #      alpha=[
                #          #          # --- 0.9354 (10/10)
                #          #          #  0.5,
                #          #          #  0.40,
                #          #          #  0.10,
                #          #          #  0.05,
                #          #          #  0.5,
                #          #          #  0.30,
                #          #          #  0.25,
                #          #          #  0.35
                #          #          # --- 0.9383 (10/10)
                #          #          0.5,
                #          #          0.40,
                #          #          0.10,
                #          #          0.05,
                #          #          0.5,
                #          #          0.40,
                #          #          0.40,
                #          #          0.40
                #          #  # focalloss_sympotom --- 0.9423 (10/10)
                #          #  # online f1: 0.92590
                #          #  0.50,  #   10
                #          #  0.40,  #   52
                #          #  0.10,  # 1267
                #          #  0.05,  # 2550
                #          #  0.50,  #    7
                #          #  0.40,  #  147
                #          #  0.35,  #  867
                #          #  0.40,  #  100
                #          #  # focalloss_sympotom --- 0.9396 (10/10)
                #          #  0.50,  #   10
                #          #  0.40,  #   52
                #          #  0.10,  # 1267
                #          #  0.05,  # 2550
                #          #  0.50,  #    7
                #          #  0.40,  #  147
                #          #  0.38,  #  867
                #          #  0.40,  #  100
                #          #  # focalloss_disease --- 0.9407 (7/10)
                #          #  # online f1: 0.92330
                #          #  0.50,  #   10
                #          #  0.40,  #   52
                #          #  0.08,  # 1267
                #          #  0.05,  # 2550
                #          #  0.50,  #    7
                #          #  0.40,  #  147
                #          #  0.35,  #  867
                #          #  0.40,  #  100
                #          #  # focalloss_others --- 0.9396 (10/10)
                #          #  0.50,  #   10
                #          #  0.40,  #   52
                #          #  0.10,  # 1267
                #          #  0.05,  # 2550
                #          #  0.50,  #    7
                #          #  0.40,  #  147
                #          #  0.35,  #  867
                #          #  0.30,  #  100
                #          # focalloss_checks --- 0.9397 (8/10)
                #          #  # online f1: 0.92640
                #          0.50,  #   10
                #          0.40,  #   52
                #          0.10,  # 1267
                #          0.05,  # 2550
                #          0.50,  #    7
                #          0.38,  #  147
                #          0.35,  #  867
                #          0.40,  #  100
                #      ])(logits.view(-1, self.num_labels), labels.view(-1))


#

#  glue_labels = ['病毒', '细菌', '疾病', '药物', '医学专科', '检查科目', '症状', 'NoneType']
# 10,     52,    1267,  2550,  7,     147,   867,   100
# [0.200, 0.200, 0.025, 0.025, 0.200, 0.100, 0.050, 0.200]

#  from ...losses import CircleLoss
#  loss = CircleLoss(m=0.25,
#                    gamma=128)
#  loss = CircleLoss(scale=32, margin=0.25)(logits.view(
#      -1, self.num_labels), labels.view(-1))
#  outputs = (loss, ) + outputs
            return (loss, ) + outputs
        else:
            # (loss), logits, (hidden_states), (attentions)
            return (torch.tensor(0.0).cuda(), ) + outputs
            #  return (None, ) + outputs
            #  return outputs

MODEL_CLASSES = {
    #  'bert': (BertConfig, BertForSequenceClassification, EmotionTokenizer),
    'bert': (BertConfig, BertForSequenceClassification, BertTokenizerFast),
    #  'albert':
    #  (AlbertConfig, AlBertForSequenceClassification, BertTokenizerFast),
}


def load_pretrained_tokenizer(args):
    #  config_class, model_class, tokenizer_class = MODEL_CLASSES[args.model_type]
    #  tokenizer = tokenizer_class.from_pretrained(
    #      args.model_path,
    #      do_lower_case=args.do_lower_case,
    #      is_english=args.is_english,
    #      emotion_words_file=args.emotion_words_file,
    #      cache_dir=args.cache_dir if args.cache_dir else None,
    #  )
    from ..token_utils import HFTokenizer
    tokenizer = HFTokenizer(os.path.join(args.model_path, 'vocab.txt'),
                            lowercase=args.do_lower_case,
                            cc=args.cc)

    return tokenizer


def load_pretrained_model(args):
    # make sure only the first process in distributed training
    # will download model & vocab
    barrier_member_processes(args)

    config_class, model_class, tokenizer_class = MODEL_CLASSES[args.model_type]
    config = config_class.from_pretrained(
        args.model_path,
        num_labels=args.num_labels,
        cache_dir=args.cache_dir if args.cache_dir else None,
    )
    setattr(config, 'loss_type', args.loss_type)
    setattr(config, 'focalloss_gamma', args.focalloss_gamma)
    setattr(config, 'focalloss_alpha', args.focalloss_alpha)
    setattr(config, 'diceloss_weight', args.diceloss_weight)

    logger.info(f"model_path: {args.model_path}")
    logger.info(f"config:{config}")
    model = model_class.from_pretrained(
        args.model_path,
        from_tf=bool(".ckpt" in args.model_path),
        config=config,
        cache_dir=args.cache_dir if args.cache_dir else None)

    # make sure only the first process in distributed training
    # will download model & vocab
    barrier_leader_process(args)

    return model


def load_model(args):
    model = load_pretrained_model(args)
    model.to(args.device)
    return model


def build_default_model(args):
    """
    自定义模型
    规格要求返回模型(model)、优化器(optimizer)、调度器(scheduler)三元组。
    """

    # -------- model --------
    model = load_pretrained_model(args)
    model.to(args.device)

    # -------- optimizer --------
    from transformers.optimization import AdamW
    optimizer_parameters = get_default_optimizer_parameters(
        model, args.weight_decay)
    optimizer = AdamW(optimizer_parameters,
                      lr=args.learning_rate,
                      correct_bias=False)

    # -------- scheduler --------
    from transformers.optimization import get_linear_schedule_with_warmup

    scheduler = get_linear_schedule_with_warmup(
        optimizer,
        num_warmup_steps=args.total_steps * args.warmup_rate,
        num_training_steps=args.total_steps)

    return model, optimizer, scheduler


def collate_fn(batch):
    """
    batch should be a list of (sequence, target, length) tuples...
    Returns a padded tensor of sequences sorted from longest to shortest,
    """
    #  all_input_ids, all_attention_mask, all_token_type_ids, all_lens, all_labels = map(
    #      torch.stack, zip(*batch))

    all_input_ids = torch.stack([e.input_ids for e in batch])
    all_attention_mask = torch.stack([e.attention_mask for e in batch])
    all_token_type_ids = torch.stack([e.token_type_ids for e in batch])
    all_input_lens = torch.stack([e.input_len for e in batch])
    all_token_offsets = torch.stack([e.token_offsets for e in batch])

    all_labels = torch.stack([e.label for e in batch])

    max_len = max(all_input_lens).item()
    all_input_ids = all_input_ids[:, :max_len]
    all_attention_mask = all_attention_mask[:, :max_len]
    all_token_type_ids = all_token_type_ids[:, :max_len]

    return all_input_ids, all_attention_mask, all_token_type_ids, all_input_lens, all_token_offsets, all_labels


class ClassReporter(object):
    def __init__(self, target_names=None, labels=None):
        logger.debug(f"target_names: {target_names}")
        logger.debug(f"labels: {labels}")
        self.target_names = target_names
        self.labels = labels

    def __call__(self, output, target):
        #  _, y_pred = torch.max(output.data, 1)
        #  y_pred = y_pred.cpu().numpy()
        #  y_true = target.cpu().numpy()

        #  y_pred = output.cpu().numpy()
        #  y_true = target.cpu().numpy()

        y_pred = output
        y_true = target

        #  print(y_pred, y_true, self.labels, self.target_names)
        #  if len(y_true.shape) == 1:
        classify_report = classification_report(y_true,
                                                y_pred,
                                                labels=self.labels,
                                                target_names=self.target_names)
        logger.info(f"classify_report:\n{classify_report}")
        #  else:
        #      for i, label in enumerate(self.labels):
        #          auc = roc_auc_score(y_score=y_pred[:, i], y_true=y_true[:, i])
        #          print(f"{label} - auc: {auc:.4f}")

        return classify_report


def init_labels(args, labels):
    args.id2label = {i: label for i, label in enumerate(labels)}
    args.label2id = {label: i for i, label in enumerate(labels)}
    args.num_labels = len(args.label2id)

    logger.info(f"args.label2id: {args.label2id}")
    logger.info(f"args.id2label: {args.id2label}")
    logger.info(f"args.num_labels: {args.num_labels}")


class GlueTrainer(Trainer):
    def __init__(self, args, glue_labels, build_model=None, tokenizer=None):
        super(GlueTrainer, self).__init__(args)
        init_labels(args, glue_labels)

        if tokenizer:
            self.tokenizer = tokenizer
        else:
            self.tokenizer = load_pretrained_tokenizer(args)

        if build_model is None:
            self.build_model = build_default_model
        else:
            self.build_model = build_model

        self.label2id = args.label2id
        self.collate_fn = collate_fn

        self.logits = None
        self.preds = None
        self.probs = None

        self.class_reporter = ClassReporter(
            target_names=[x for x in self.label2id.keys()],
            labels=[x for x in self.label2id.values()])

    #  def generate_dataloader(self, args, dataset, batch_size, keep_order=True):
    #
    #      Sampler = SequentialSampler if keep_order else RandomSampler
    #      sampler = DistributedSampler(dataset) if is_multi_processes(
    #          args) else Sampler(dataset)
    #      dataloader = DataLoader(dataset,
    #                              sampler=sampler,
    #                              batch_size=batch_size,
    #                              collate_fn=collate_fn)
    #      return dataloader

    #  def examples_to_dataset(self, examples, max_seq_length):
    #      from .dataset import examples_to_dataset
    #      return examples_to_dataset(examples, self.label2id, self.tokenizer,
    #                                 max_seq_length)
    def encode_examples(self, examples, max_seq_length):
        from .dataset import encode_examples
        return encode_examples(examples, self.label2id, self.tokenizer,
                               max_seq_length)

    def batch_to_inputs(self, args, batch, known_labels=True):
        inputs = {
            "input_ids": batch[0],
            "attention_mask": batch[1],
            "labels": batch[5]
        }
        if args.model_type != "distilbert":
            # XLM, DistilBERT, RoBERTa, and XLM-RoBERTa don't use segment_ids
            inputs["token_type_ids"] = (batch[2] if args.model_type in [
                "bert", "xlnet", "albert"
            ] else None)
        return inputs

    def on_eval_start(self, args, eval_dataset):
        self.logits = None
        self.preds = None
        self.out_label_ids = None
        self.results = {}
        self.probs = None

    #  def on_eval_step(self, args, eval_dataset, step, model, inputs, outputs):
    def on_eval_step(self, args, model, step, batch):
        input_ids, attention_mask, _, input_lens, token_offsets, label_ids = batch
        outputs = self.on_train_step(args, model, step, batch)
        eval_loss, logits = outputs[:2]
        if self.logits is None:
            self.logits = logits.detach().cpu().numpy()
            self.out_label_ids = label_ids.detach().cpu().numpy()
        else:
            self.logits = np.append(self.logits,
                                    logits.detach().cpu().numpy(),
                                    axis=0)
            self.out_label_ids = np.append(self.out_label_ids,
                                           label_ids.detach().cpu().numpy(),
                                           axis=0)

        self.preds = np.argmax(self.logits, axis=1)
        self.probs = softmax(self.logits)

        # 调整类别0的判断阈值
        #  self.preds = np.array([
        #      0 if x == 1 and prob[1] < 0.60 else x
        #      for prob, x in zip(self.probs, self.preds)
        #  ])
        self.preds, probs = logits_to_preds(self.logits)
        #  logger.debug(f"self.preds: {self.preds.shape}, {self.preds}")
        #  logger.debug(
        #      f"self.out_label_ids: {self.out_label_ids.shape}, {self.out_label_ids}, {np.argmax(self.out_label_ids, axis=1)}"
        #  )

        if len(self.out_label_ids.shape) == 1:
            result = acc_and_f1(self.preds, self.out_label_ids)
        else:
            #  result = acc_and_f1(self.preds,
            #                      np.argmax(self.out_label_ids, axis=1))
            self.preds = np.array([
                x > args.confidence for x in sigmoid(self.logits)
            ]).astype(int)
            result = acc_and_f1(self.preds, self.out_label_ids)

        #  if args.do_experiment:
        #      #  mlflow.log_metric('loss', eval_loss.item())
        #      for key, value in result.items():
        #          mlflow.log_metric(key, value)

        self.results.update(result)

        return (eval_loss, ), self.results

    def on_eval_end(self, args, eval_dataset):
        if len(self.out_label_ids.shape) == 1:
            self.preds = np.argmax(self.logits, axis=1)
        else:
            self.preds = np.array([
                x > args.confidence for x in sigmoid(self.logits)
            ]).astype(int)
        # for regessions
        #  self.preds = np.squeeze(self.preds)

        #  logger.debug(f"self.preds: {self.preds.shape}, {self.preds}")
        #  logger.debug(
        #      f"self.out_label_ids: {self.out_label_ids.shape}, {self.out_label_ids}"
        #  )
        self.class_reporter(self.preds, self.out_label_ids)

        logger.info(f"  Num examples = {len(eval_dataset)}")
        logger.info(f"  Batch size = {args.eval_batch_size}")
        logger.info(f"******** Eval results ********")
        for key in self.results.keys():
            logger.info(f" dev: {key} = {self.results[key]:.6f}")

        return self.results

    def on_predict_start(self, args, test_dataset):
        self.logits = None

        self.pred_results = None
        self.pred_probs = None

    #  def on_predict_step(self, args, test_dataset, step, model, inputs,
    #                      outputs):
    def on_predict_step(self, args, model, step, batch):
        inputs = self.batch_to_inputs(args, batch, known_labels=False)
        outputs = model(**inputs)

        #  logger.debug(f"inputs: {inputs}")
        #  logger.debug(f"outputs: {outputs}")
        logits = outputs[1]
        #  logits = outputs[0]
        if self.logits is None:
            self.logits = logits.detach().cpu().numpy()
        else:
            self.logits = np.append(self.logits,
                                    logits.detach().cpu().numpy(),
                                    axis=0)

    def on_predict_end(self, args, test_dataset):
        logger.warning(f"self.logits.shape: {self.logits.shape}")
        self.pred_results = np.argmax(self.logits, axis=1)
        self.pred_probs = softmax(self.logits)

        #  self.pred_results = np.array([
        #      0 if x == 1 and prob[1] < 0.60 else x
        #      for prob, x in zip(self.pred_probs, self.pred_results)
        #  ])

        # multi labels
        #  threshold = 5.0 / len(self.label2id)
        #  logger.info(f"self.pred_probs > threshold: {self.pred_probs > threshold}")
        #  logger.info(
        #      f"np.sum(self.pred_probs > threshold: {np.sum(self.pred_probs > threshold)}")
        #  threshold = args.confidence
        #  self.pred_probs = sigmoid(self.logits)
        #  self.pred_results = np.array(self.pred_probs > threshold, dtype=np.int64)

        # multi classes
        #  self.pred_results, self.pred_probs = logits_to_preds(self.logits)

        logger.debug(f"pred_results: {self.pred_results}")
        logger.debug(f"pred_probs: {self.pred_probs}")



# -*- coding: utf-8 -*-
# File: stats.py

import numpy as np

__all__ = ['StatCounter', 'BinaryStatistics', 'RatioCounter', 'Accuracy',
           'OnlineMoments']


class StatCounter(object):
    """ A simple counter"""

    def __init__(self):
        self.reset()

    def feed(self, v):
        """
        Args:
            v(float or np.ndarray): has to be the same shape between calls.
        """
        self._values.append(v)

    def reset(self):
        self._values = []

    @property
    def count(self):
        return len(self._values)

    @property
    def average(self):
        assert len(self._values)
        return np.mean(self._values)

    @property
    def sum(self):
        assert len(self._values)
        return np.sum(self._values)

    @property
    def max(self):
        assert len(self._values)
        return max(self._values)

    @property
    def min(self):
        assert len(self._values)
        return min(self._values)

    def samples(self):
        """
        Returns all samples.
        """
        return self._values


class RatioCounter(object):
    """ A counter to count ratio of something. """

    def __init__(self):
        self.reset()

    def reset(self):
        self._tot = 0
        self._cnt = 0

    def feed(self, count, total=1):
        """
        Args:
            cnt(int): the count of some event of interest.
            tot(int): the total number of events.
        """
        self._tot += total
        self._cnt += count

    @property
    def ratio(self):
        if self._tot == 0:
            return 0
        return self._cnt * 1.0 / self._tot

    @property
    def total(self):
        """
        Returns:
            int: the total
        """
        return self._tot

    @property
    def count(self):
        """
        Returns:
            int: the total
        """
        return self._cnt


class Accuracy(RatioCounter):
    """ A RatioCounter with a fancy name """
    @property
    def accuracy(self):
        return self.ratio


class BinaryStatistics(object):
    """
    Statistics for binary decision,
    including precision, recall, false positive, false negative
    """

    def __init__(self):
        self.reset()

    def reset(self):
        self.nr_pos = 0  # positive label
        self.nr_neg = 0  # negative label
        self.nr_pred_pos = 0
        self.nr_pred_neg = 0
        self.corr_pos = 0   # correct predict positive
        self.corr_neg = 0   # correct predict negative

    def feed(self, pred, label):
        """
        Args:
            pred (np.ndarray): binary array.
            label (np.ndarray): binary array of the same size.
        """
        assert pred.shape == label.shape, "{} != {}".format(pred.shape, label.shape)
        self.nr_pos += (label == 1).sum()
        self.nr_neg += (label == 0).sum()
        self.nr_pred_pos += (pred == 1).sum()
        self.nr_pred_neg += (pred == 0).sum()
        self.corr_pos += ((pred == 1) & (pred == label)).sum()
        self.corr_neg += ((pred == 0) & (pred == label)).sum()

    @property
    def precision(self):
        if self.nr_pred_pos == 0:
            return 0
        return self.corr_pos * 1. / self.nr_pred_pos

    @property
    def recall(self):
        if self.nr_pos == 0:
            return 0
        return self.corr_pos * 1. / self.nr_pos

    @property
    def false_positive(self):
        if self.nr_pred_pos == 0:
            return 0
        return 1 - self.precision

    @property
    def false_negative(self):
        if self.nr_pos == 0:
            return 0
        return 1 - self.recall


class OnlineMoments(object):
    """Compute 1st and 2nd moments online (to avoid storing all elements).

    See algorithm at: https://www.wikiwand.com/en/Algorithms_for_calculating_variance#/Online_algorithm
    """

    def __init__(self):
        self._mean = 0
        self._M2 = 0
        self._n = 0

    def feed(self, x):
        """
        Args:
            x (float or np.ndarray): must have the same shape.
        """
        self._n += 1
        delta = x - self._mean
        self._mean += delta * (1.0 / self._n)
        delta2 = x - self._mean
        self._M2 += delta * delta2

    @property
    def mean(self):
        return self._mean

    @property
    def variance(self):
        return self._M2 / (self._n - 1)

    @property
    def std(self):
        return np.sqrt(self.variance)



# -*- coding: utf-8 -*-
# File: mnist.py


import gzip
import numpy
import os

from ...utils import logger
from ...utils.fs import download, get_dataset_path
from ..base import RNGDataFlow

__all__ = ['Mnist', 'FashionMnist']


def maybe_download(url, work_directory):
    """Download the data from Yann's website, unless it's already here."""
    filename = url.split('/')[-1]
    filepath = os.path.join(work_directory, filename)
    if not os.path.exists(filepath):
        logger.info("Downloading to {}...".format(filepath))
        download(url, work_directory)
    return filepath


def _read32(bytestream):
    dt = numpy.dtype(numpy.uint32).newbyteorder('>')
    return numpy.frombuffer(bytestream.read(4), dtype=dt)[0]


def extract_images(filename):
    """Extract the images into a 4D uint8 numpy array [index, y, x, depth]."""
    with gzip.open(filename) as bytestream:
        magic = _read32(bytestream)
        if magic != 2051:
            raise ValueError(
                'Invalid magic number %d in MNIST image file: %s' %
                (magic, filename))
        num_images = _read32(bytestream)
        rows = _read32(bytestream)
        cols = _read32(bytestream)
        buf = bytestream.read(rows * cols * num_images)
        data = numpy.frombuffer(buf, dtype=numpy.uint8)
        data = data.reshape(num_images, rows, cols, 1)
        data = data.astype('float32') / 255.0
        return data


def extract_labels(filename):
    """Extract the labels into a 1D uint8 numpy array [index]."""
    with gzip.open(filename) as bytestream:
        magic = _read32(bytestream)
        if magic != 2049:
            raise ValueError(
                'Invalid magic number %d in MNIST label file: %s' %
                (magic, filename))
        num_items = _read32(bytestream)
        buf = bytestream.read(num_items)
        labels = numpy.frombuffer(buf, dtype=numpy.uint8)
        return labels


class Mnist(RNGDataFlow):
    """
    Produces [image, label] in MNIST dataset,
    image is 28x28 in the range [0,1], label is an int.
    """

    _DIR_NAME = 'mnist_data'
    _SOURCE_URL = 'http://yann.lecun.com/exdb/mnist/'

    def __init__(self, train_or_test, shuffle=True, dir=None):
        """
        Args:
            train_or_test (str): either 'train' or 'test'
            shuffle (bool): shuffle the dataset
        """
        if dir is None:
            dir = get_dataset_path(self._DIR_NAME)
        assert train_or_test in ['train', 'test']
        self.train_or_test = train_or_test
        self.shuffle = shuffle

        def get_images_and_labels(image_file, label_file):
            f = maybe_download(self._SOURCE_URL + image_file, dir)
            images = extract_images(f)
            f = maybe_download(self._SOURCE_URL + label_file, dir)
            labels = extract_labels(f)
            assert images.shape[0] == labels.shape[0]
            return images, labels

        if self.train_or_test == 'train':
            self.images, self.labels = get_images_and_labels(
                'train-images-idx3-ubyte.gz',
                'train-labels-idx1-ubyte.gz')
        else:
            self.images, self.labels = get_images_and_labels(
                't10k-images-idx3-ubyte.gz',
                't10k-labels-idx1-ubyte.gz')

    def __len__(self):
        return self.images.shape[0]

    def __iter__(self):
        idxs = list(range(self.__len__()))
        if self.shuffle:
            self.rng.shuffle(idxs)
        for k in idxs:
            img = self.images[k].reshape((28, 28))
            label = self.labels[k]
            yield [img, label]


class FashionMnist(Mnist):
    """
    Same API as :class:`Mnist`, but more fashion.
    """

    _DIR_NAME = 'fashion_mnist_data'
    _SOURCE_URL = 'http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/'

    def get_label_names(self):
        """
        Returns:
            [str]: the name of each class
        """
        # copied from https://github.com/zalandoresearch/fashion-mnist
        return ['T-shirt/top', 'Trouser', 'Pullover', 'Dress', 'Coat',
                'Sandal', 'Shirt', 'Sneaker', 'Bag', 'Ankle boot']


if __name__ == '__main__':
    ds = Mnist('train')
    ds.reset_state()
    for _ in ds:
        from IPython import embed
        embed()
        break



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os, sys, json
import glob, shutil
from datetime import datetime

import rich
from loguru import logger
#  from rich import print
from rich.console import Console
from tqdm import tqdm

from .modeling import NerDataset, SpoDataset

console = Console()

all_models = []


def find_models(args):
    global all_models

    output_dir = args.output_dir
    files = glob.glob(os.path.join(output_dir, "*/local_id"))
    files += glob.glob(os.path.join(output_dir, "saved_models", "*/local_id"))

    for file in files:
        local_id = None
        with open(file, 'r') as rd:
            local_id = rd.readline().strip()
            model_path = os.path.split(file)[0]

            ctime = os.stat(file).st_ctime
            ctime = datetime.fromtimestamp(ctime).strftime('%Y/%m/%d %H:%M:%S')

            all_models.append((local_id, model_path, ctime))
    all_models = sorted(all_models, key=lambda x: x[2], reverse=True)


def get_model(local_id):
    for model in all_models:
        if model[0] == local_id:
            return model
    return None


def get_model_args_path(model):
    if model is None:
        args_path = os.path.join(f"{args.output_dir}",
                                 "latest/best/training_args.json")
        if not os.path.exists(args_path):
            args_path = os.path.join(f"{args.output_dir}",
                                     "latest/training_args.json")

    else:
        args_path = os.path.join(model[1], "best/training_args.json")
        if not os.path.exists(args_path):
            args_path = os.path.join(model[1], "training_args.json")

    return args_path


def show_model(args):
    logger.info(f"{args.local_id}")
    model = None
    if args.local_id:
        model_id = args.local_id[0]
        model = get_model(model_id)
    #  for model_id in args.local_id:
    #      model = get_model(model_id)

    args_path = get_model_args_path(model)

    ctime = os.stat(args_path).st_ctime
    ctime = datetime.fromtimestamp(ctime).strftime('%Y/%m/%d %H:%M:%S')

    training_args = json.load(open(args_path))
    logger.warning(f"-------- {training_args['local_id']} --------")
    logger.info(f'{ctime}')

    logger.info(f"dataset_name: {training_args['dataset_name']}")
    logger.info(f"num_train_epochs: {training_args['num_train_epochs']}")
    logger.info(f"learning_rate: {training_args['learning_rate']}")
    logger.info(
        f"train_max_seq_length: {training_args['train_max_seq_length']}")
    logger.info(f"eval_max_seq_length: {training_args['eval_max_seq_length']}")
    logger.info(
        f"per_gpu_train_batch_size: {training_args['per_gpu_train_batch_size']}"
    )
    logger.info(
        f"per_gpu_eval_batch_size: {training_args['per_gpu_eval_batch_size']}")
    logger.info(
        f"per_gpu_predict_batch_size: {training_args['per_gpu_predict_batch_size']}"
    )

    logger.info('-' * 50)
    logger.debug(f"local_dir: {training_args['local_dir']}")
    logger.debug(f"train_file: {training_args['train_file']}")
    logger.debug(f"eval_file: {training_args['eval_file']}")
    logger.debug(f"test_file: {training_args['test_file']}")

    logger.info(f"fold: {training_args['fold']}")
    logger.info(f"num_augments: {training_args['num_augments']}")
    logger.info(f"seg_len: {training_args['seg_len']}")
    logger.info(f"seg_backoff: {training_args['seg_backoff']}")
    logger.info(f"train_rate: {training_args['train_rate']}")
    logger.info(f"train_sample_rate: {training_args['train_sample_rate']}")
    max_train_examples = training_args.get('max_train_examples', 0)
    if max_train_examples > 0:
        logger.warning(f"max_train_examples: {max_train_examples}")
    else:
        logger.debug(f"max_train_examples: {max_train_examples}")
    logger.info(f"random_type: {training_args.get('random_type', None)}")
    confidence = training_args.get('confidence', 0.5)
    logger.info(f"confidence: {confidence}")
    enable_nested_entities = training_args.get('enable_nested_entities', False)
    if enable_nested_entities:
        logger.warning(f"enable_nested_entities: {enable_nested_entities}")
    else:
        logger.info(f"enable_nested_entities: {enable_nested_entities}")

    logger.info('-' * 50)

    enable_fp16 = training_args.get('fp16', None)
    if enable_fp16:
        logger.warning(f"fp16: {enable_fp16}")
    else:
        logger.debug(f"fp16: {enable_fp16}")

    enable_kd = training_args.get('enable_kd', None)
    if enable_kd:
        logger.warning(f"enable_kd: {enable_kd}")
    else:
        logger.debug(f"enable_kd: {enable_kd}")

    enable_sda = training_args.get('enable_sda', None)
    if enable_sda:
        logger.warning(f"enable_sda: {enable_sda}")
        logger.info(f"sda_teachers: {training_args['sda_teachers']}")
        logger.info(f"sda_coeff: {training_args['sda_coeff']:.2f}")
        logger.info(f"sda_decay: {training_args['sda_decay']:.3f}")
    else:
        logger.debug(f"enable_sda: {enable_sda}")

    logger.info('-' * 50)

    ner_type = training_args.get("ner_type", None)
    if ner_type:
        logger.info(f"ner_type: {training_args['ner_type']}")
    logger.info(f"model_type: {training_args['model_type']}")
    logger.info(f"model_path: {training_args['model_path']}")
    logger.info(f"tracking_uri: {training_args['tracking_uri']}")
    logger.info(f"num_labels: {training_args['num_labels']}")
    logger.info(f"seed: {training_args['seed']}")
    logger.info(f"loss_type: {training_args['loss_type']}")
    if training_args['loss_type'] == 'FocalLoss':
        logger.info(f"focalloss_gamma: {training_args['focalloss_gamma']}")
        logger.info(f"focalloss_alpha: {training_args['focalloss_alpha']}")

    if args.detail:
        logger.warning('-' * 50)
        logger.info(
            json.dumps({k: v
                        for k, v in sorted(training_args.items())},
                       ensure_ascii=False,
                       indent=2))


def list_models(args):
    print('-' * 102)
    print("local_id", ' ' * 28, "ctime", ' ' * 15, "model_path")
    print('-' * 102)
    for local_id, model_path, ctime in all_models:
        print(local_id, '    ', ctime, ' ', model_path)


def diff_models(args):
    logger.info(f"{args.local_id}")
    if len(args.local_id) >= 1:

        if len(args.local_id) >= 2:
            src_model_id = args.local_id[0]
            src_model = get_model(src_model_id)
            if src_model is None:
                logger.warning(f"Model {src_model_id} not found.")
                return
            dest_model_id = args.local_id[1]
            dest_model = get_model(dest_model_id)
            if dest_model is None:
                logger.warning(f"Model {dest_model_id} not found.")
                return

            logger.info(f"{[src_model[2], dest_model[2]]}")
        else:
            src_model = None
            dest_model_id = args.local_id[0]
            dest_model = get_model(dest_model_id)
            if dest_model is None:
                logger.warning(f"Model {dest_model_id} not found.")
                return

        #  src_args_path = os.path.join(src_model[1], "best/training_args.json")
        #  if not os.path.exists(src_args_path):
        #      src_args_path = os.path.join(src_model[1], "training_args.json")
        #  dest_args_path = os.path.join(dest_model[1], "best/training_args.json")
        #  if not os.path.exists(dest_args_path):
        #      dest_args_path = os.path.join(dest_model[1], "training_args.json")

        src_args_path = get_model_args_path(src_model)
        dest_args_path = get_model_args_path(dest_model)

        src_args = json.load(open(src_args_path))
        dest_args = json.load(open(dest_args_path))

        for k, v in sorted(src_args.items()):
            if k in dest_args and v == dest_args[k]:
                continue
            logger.debug(f"{k}, {v}")
            logger.debug(f"{k}, {dest_args.get(k, None)}")
            logger.debug('')

        for k, v in sorted(dest_args.items()):
            if k not in src_args:
                logger.debug(f"{k}, {src_args.get(k, None)}")
                logger.debug(f"{k}, {v}")
                logger.debug('')


def new_model(args):
    latest_dir = os.path.join(args.output_dir, "latest")
    if os.path.exists(latest_dir):
        shutil.rmtree(latest_dir)
    os.makedirs(latest_dir)

    theta_src_path = get_theta_src_path()
    logger.warning(f"theta_src_path: {theta_src_path}")
    if theta_src_path:
        app_type = args.app_type
        dataset_name = args.dataset_name
        cmd_cp_params = f"cp {theta_src_path}/templates/{app_type}/{app_type}_params.py {latest_dir}/{dataset_name}_params.py"
        logger.warning(f"{cmd_cp_params}")
        os.system(cmd_cp_params)

    import uuid
    local_id = str(uuid.uuid1()).replace('-', '')[:8]
    local_id_file = os.path.join(latest_dir, "local_id")
    with open(local_id_file, 'w') as wt:
        wt.write(f"{local_id}")
    logger.info(f"New model id: {local_id}")


def use_model(args):
    logger.info(f"{args.local_id}")
    if len(args.local_id) >= 1:
        model_id = args.local_id[0]
        model = get_model(model_id)
        if model:
            local_id, model_path, ctime = model

            latest_dir = os.path.join(args.output_dir, "latest")
            if os.path.exists(latest_dir):
                shutil.rmtree(latest_dir)
            shutil.copytree(model_path, latest_dir)
            logger.info(
                f"Use local model({local_id}) {model_path} to {latest_dir}")


def get_dataset_name(args):
    dataset_name = args.dataset_name
    if dataset_name is None:
        dataset_name = os.path.basename(os.path.abspath(os.curdir))
    return dataset_name


def get_dataset_module(dataset_name):
    import importlib
    dataset_module = importlib.import_module(f"{dataset_name}")

    return dataset_module


def export_train_data(args):
    dataset_name = get_dataset_name(args)
    dataset_module = get_dataset_module(dataset_name)

    train_data_generator = None
    if 'train_data_generator' in dataset_module.__dict__:
        train_data_generator = dataset_module.train_data_generator
    ner_labels = None
    if 'ner_labels' in dataset_module.__dict__:
        ner_labels = dataset_module.ner_labels
    ner_connections = None
    if 'ner_connections' in dataset_module.__dict__:
        ner_connections = dataset_module.ner_connections
    glue_labels = None
    if 'glue_labels' in dataset_module.__dict__:
        glue_labels = dataset_module.glue_labels

    predicate_labels = None
    if 'predicate_labels' in dataset_module.__dict__:
        predicate_labels = dataset_module.predicate_labels

    if predicate_labels:
        console.log("[bold cyan]predicate_labels:[/bold cyan]",
                    predicate_labels)
        spo_dataset = SpoDataset(dataset_name, predicate_labels)
        spo_dataset.load(train_data_generator)
        spo_dataset.info()
        export_format = args.format
        if export_format == "json":
            dataset_file = args.dataset_file
            spo_dataset.save(dataset_file)
        else:
            raise Exception(
                f"Bad export format {export_format}. Only available ['json']")
    elif ner_labels:
        console.log("[bold cyan]ner_labels:[/bold cyan]", ner_labels)
        console.log("[bold cyan]ner_connections:[/bold cyan]", ner_connections)

        ner_dataset = NerDataset(dataset_name, ner_labels, ner_connections)
        dataset_file = args.dataset_file
        console.log("[bold cyan]dataset_file:[/bold cyan]", dataset_file)
        if dataset_file is not None:
            ner_dataset.load_from_file(dataset_file)
        else:
            ner_dataset.load(train_data_generator)
        ner_dataset.info()

        export_format = args.format
        if export_format == "json" or export_format == "lines":
            dataset_file = args.output
            if not dataset_file:
                dataset_file = f"{dataset_name}_train_data.json"
            ner_dataset.save(dataset_file, format=export_format)
        elif export_format == "brat":
            brat_data_dir = args.brat_data_dir
            if not brat_data_dir:
                brat_data_dir = "brat_data"
            if not os.path.exists(brat_data_dir):
                os.makedirs(brat_data_dir)
            ner_dataset.export_to_brat(brat_data_dir, max_pages=args.max_pages)
        elif export_format == "poplar":
            pass
        else:
            raise Exception(
                f"Bad export format {export_format}. Only available ['json', 'brat', 'poplar']"
            )
    elif glue_labels:
        console.log("[bold cyan]glue_labels:[/bold cyan]", glue_labels)

        export_format = args.format
        dataset_file = args.output
        if not dataset_file:
            dataset_file = f"{dataset_name}_train_data.json"
        all_examples = []
        for guid, text, _, labels in tqdm(train_data_generator()):
            all_examples.append({'guid': guid, 'text': text, 'labels': labels})
        json.dump(all_examples,
                  open(dataset_file, 'w'),
                  ensure_ascii=False,
                  indent=2)
        logger.warning(f"Saved {dataset_file}")


def reviews_data_generator(reviews_file):
    json_data = json.load(open(reviews_file, 'r'))
    for guid, items in json_data.items():
        text = items['text']
        entities = items['tags']
        tags = []
        for ent in entities:
            tags.append({
                'category': ent['category'],
                'start': ent['start'],
                'mention': ent['mention']
            })
        yield guid, text, None, tags


def export_test_data(args):
    dataset_name = get_dataset_name(args)
    dataset_module = get_dataset_module(dataset_name)

    test_data_generator = None
    if 'test_data_generator' in dataset_module.__dict__:
        test_data_generator = dataset_module.test_data_generator
    ner_labels = None
    if 'ner_labels' in dataset_module.__dict__:
        ner_labels = dataset_module.ner_labels
    ner_connections = None
    if 'ner_connections' in dataset_module.__dict__:
        ner_connections = dataset_module.ner_connections
    glue_labels = None
    if 'glue_labels' in dataset_module.__dict__:
        glue_labels = dataset_module.glue_labels

    #  logger.info(f"ner_labels: {ner_labels}")
    #  logger.info(f"ner_connections: {ner_connections}")
    console.log("[bold cyan]ner_labels:[/bold cyan]", ner_labels)
    console.log("[bold cyan]ner_connections:[/bold cyan]", ner_connections)

    if args.local_id:
        local_id = args.local_id[0]
    else:
        args_path = get_model_args_path(None)
        training_args = json.load(open(args_path))
        local_id = training_args['local_id']

    if ner_labels:
        reviews_file = os.path.join(args.output_dir, 'saved_models', local_id,
                                    f"{dataset_name}_reviews_{local_id}.json")

        ner_dataset = NerDataset(dataset_name, ner_labels, ner_connections)
        ner_dataset.load(reviews_data_generator, reviews_file)
        ner_dataset.info()

        export_format = args.format
        if export_format == "json" or export_format == "lines":
            dataset_file = args.output
            if not dataset_file:
                dataset_file = f"{dataset_name}_test_data.json"
            ner_dataset.save(dataset_file, format=export_format)
        elif export_format == "brat":
            brat_data_dir = args.brat_data_dir
            if not brat_data_dir:
                brat_data_dir = "brat_data"
            if not os.path.exists(brat_data_dir):
                os.makedirs(brat_data_dir)
            ner_dataset.export_to_brat(brat_data_dir, max_pages=args.max_pages)
        elif export_format == "poplar":
            pass
        else:
            raise Exception(
                f"Bad export format {export_format}. Only available ['json', 'brat', 'poplar']"
            )
    elif glue_labels:
        console.log("[bold cyan]glue_labels:[/bold cyan]", glue_labels)

        export_format = args.format
        dataset_file = args.output
        if not dataset_file:
            dataset_file = f"{dataset_name}_train_data.json"
        all_examples = []
        for guid, text, _, _ in tqdm(test_data_generator()):
            all_examples.append({'guid': guid, 'text': text})
        json.dump(all_examples,
                  open(dataset_file, 'w'),
                  ensure_ascii=False,
                  indent=2)
        logger.warning(f"Saved {dataset_file}")


def export_submit_data(args):
    dataset_name = get_dataset_name(args)
    dataset_module = get_dataset_module(dataset_name)

    test_data_generator = None
    if 'test_data_generator' in dataset_module.__dict__:
        test_data_generator = dataset_module.test_data_generator
    ner_labels = None
    if 'ner_labels' in dataset_module.__dict__:
        ner_labels = dataset_module.ner_labels
    ner_connections = None
    if 'ner_connections' in dataset_module.__dict__:
        ner_connections = dataset_module.ner_connections
    glue_labels = None
    if 'glue_labels' in dataset_module.__dict__:
        glue_labels = dataset_module.glue_labels

    #  logger.info(f"ner_labels: {ner_labels}")
    #  logger.info(f"ner_connections: {ner_connections}")
    console.log("[bold cyan]ner_labels:[/bold cyan]", ner_labels)
    console.log("[bold cyan]ner_connections:[/bold cyan]", ner_connections)

    if args.local_id:
        local_id = args.local_id[0]
    else:
        args_path = get_model_args_path(None)
        training_args = json.load(open(args_path))
        local_id = training_args['local_id']

    if ner_labels:
        reviews_file = os.path.join(
            args.submissions_dir,
            f"{dataset_name}_final_reviews_{local_id}.json")

        ner_dataset = NerDataset(dataset_name, ner_labels, ner_connections)
        ner_dataset.load(reviews_data_generator, reviews_file)
        ner_dataset.info()

        export_format = args.format
        if export_format == "json" or export_format == "lines":
            dataset_file = args.output
            if not dataset_file:
                dataset_file = f"{dataset_name}_test_data.json"
            ner_dataset.save(dataset_file, format=export_format)
        elif export_format == "brat":
            brat_data_dir = args.brat_data_dir
            if not brat_data_dir:
                brat_data_dir = "brat_data"
            if not os.path.exists(brat_data_dir):
                os.makedirs(brat_data_dir)
            ner_dataset.export_to_brat(brat_data_dir, max_pages=args.max_pages)
        elif export_format == "poplar":
            pass
        else:
            raise Exception(
                f"Bad export format {export_format}. Only available ['json', 'brat', 'poplar']"
            )
    elif glue_labels:
        console.log("[bold cyan]glue_labels:[/bold cyan]", glue_labels)

        export_format = args.format
        dataset_file = args.output
        if not dataset_file:
            dataset_file = f"{dataset_name}_train_data.json"
        all_examples = []
        for guid, text, _, _ in tqdm(test_data_generator()):
            all_examples.append({'guid': guid, 'text': text})
        json.dump(all_examples,
                  open(dataset_file, 'w'),
                  ensure_ascii=False,
                  indent=2)
        logger.warning(f"Saved {dataset_file}")


def import_brat_data(args):
    dataset_name = get_dataset_name(args)
    brat_data_dir = args.brat_data_dir
    logger.info(
        f"Loading brat data from {brat_data_dir} into {dataset_name} dataset.")

    ner_dataset = NerDataset(dataset_name)
    ner_dataset.load_from_brat_data(brat_data_dir)
    ner_dataset.save(f"{dataset_name}_from_brat.json")


def import_poplar_data(args):
    pass


def json_to_brat(args):
    dataset_name = args.dataset_name

    dataset_module = get_dataset_module(dataset_name)

    ner_labels = None
    if 'ner_labels' in dataset_module.__dict__:
        ner_labels = dataset_module.ner_labels
    ner_connections = None
    if 'ner_connections' in dataset_module.__dict__:
        ner_connections = dataset_module.ner_connections

    dataset_file = args.dataset_file
    brat_data_dir = args.brat_data_dir

    ner_dataset = NerDataset(dataset_name, ner_labels, ner_connections)
    ner_dataset.load_from_file(dataset_file)
    ner_dataset.info()

    brat_data_dir = args.brat_data_dir
    logger.warning(f"brat_data_dir: {brat_data_dir}")
    if not brat_data_dir:
        brat_data_dir = "brat_data"
    if not os.path.exists(brat_data_dir):
        os.makedirs(brat_data_dir)
    ner_dataset.export_to_brat(brat_data_dir, max_pages=args.max_pages)


def brat_to_json(args):
    dataset_file = args.dataset_file
    brat_data_dir = args.brat_data_dir

    dataset = NerDataset()

    dataset.load_from_brat_data(brat_data_dir)
    dataset.save(dataset_file)

    logger.info(f"Brat {brat_data_dir} to {dataset_file}")


def run_brat(args):
    brat_collections_dir = os.path.abspath(args.brat_collections_dir)
    brat_port = args.brat_port

    os.system(
        f" docker run -it --rm -p {brat_port}:8001 -v {brat_collections_dir}:/brat/data/collections  ootb/brat:latest"
    )


def diff_ner_datasets(args):
    dataset_files = args.diff_ner_datasets
    assert dataset_files and len(dataset_files) == 2
    from theta.modeling import NerDataset

    ner_dataset_a = NerDataset("a")
    ner_dataset_b = NerDataset("b")

    #  ner_dataset_a.load_from_file(dataset_files[0])
    #  ner_dataset_b.load_from_file(dataset_files[1])
    ner_dataset_a.load(reviews_data_generator, dataset_files[0])
    ner_dataset_b.load(reviews_data_generator, dataset_files[1])

    identical_tags_list, a_only_tags_list, b_only_tags_list = ner_dataset_a.diff(
        ner_dataset_b)

    total_a_only = 0
    total_b_only = 0
    for example, identical_tags, a_only_tags, b_only_tags in zip(
            ner_dataset_a, identical_tags_list, a_only_tags_list,
            b_only_tags_list):
        if a_only_tags or b_only_tags:
            logger.info(f"{example.guid}: {example.text[:256]}")
            #  logger.info(f"{identical_tags}")
            logger.info(f"a_only: {a_only_tags}")
            logger.warning(f"b_only: {b_only_tags}")
            logger.info(f"")
            total_a_only += len(a_only_tags)
            total_b_only += len(b_only_tags)
    logger.warning(f"Total a_only (Removed): {total_a_only}")
    logger.warning(f"Total b_only (Added): {total_b_only}")


def merge_ner_datasets(args):
    dataset_files = args.merge_ner_datasets
    min_dups = args.min_dups
    assert dataset_files and len(dataset_files) >= 2

    ner_datasets = [NerDataset(f"{i}") for i in range(len(dataset_files))]
    for dataset, filename in zip(ner_datasets, dataset_files):
        if not os.path.exists(filename):
            local_id = filename
            guess_filename = f"{args.output_dir}/{local_id}/{args.dataset_name}_reviews_{local_id}.json"
            if not os.path.exists(guess_filename):
                raise Exception(f"NerDataset {filename} does not exists.")
            filename = guess_filename
        #  dataset.load_from_file(filename)
        logger.info(f"{filename}")
        dataset.load(reviews_data_generator, filename)

    from theta.modeling import merge_ner_datasets as do_merge_ner_datasets
    merged_dataset = do_merge_ner_datasets(ner_datasets, min_dups=min_dups)

    merged_dataset_file = args.dataset_file
    merged_dataset.save(merged_dataset_file, format='json')
    logger.info(f"Saved {merged_dataset_file}")


def mix_ner_datasets(args):
    dataset_files = args.mix_ner_datasets
    assert dataset_files and len(dataset_files) >= 2

    ner_datasets = [NerDataset(f"{i}") for i in range(len(dataset_files))]
    for dataset, filename in zip(ner_datasets, dataset_files):
        if not os.path.exists(filename):
            local_id = filename
            guess_filename = f"{args.output_dir}/{local_id}/{args.dataset_name}_reviews_{local_id}.json"
            if not os.path.exists(guess_filename):
                raise Exception(f"NerDataset {filename} does not exists.")
            filename = guess_filename
        #  dataset.load_from_file(filename)
        logger.info(f"{filename}")
        dataset.load(reviews_data_generator, filename)

    from theta.modeling import mix_ner_datasets as do_mix_ner_datasets
    mixed_dataset = do_mix_ner_datasets(ner_datasets)

    mixed_dataset_file = args.dataset_file
    mixed_dataset.save(mixed_dataset_file, format='json')
    logger.info(f"Saved {mixed_dataset_file}")


def get_theta_src_path():
    # -------- 检查THETA_SRC环工--------
    if not "THETA_SRC" in os.environ:
        logger.error(
            f"THETA_SRC environment variable must be set to theta src path.")
        return None
    theta_src_path = os.environ['THETA_SRC']
    if not os.path.exists(os.path.join(theta_src_path, "__init__.py")):
        logger.error(
            f"Ensure theta source must be located in THETA_SRC({theta_src_path})"
        )
        return None

    return theta_src_path


def init(args):
    theta_src_path = get_theta_src_path()
    if theta_src_path:
        app_type = args.app_type
        dataset_name = args.dataset_name
        os.system(f"cp {theta_src_path}/templates/{app_type}/* .")
        os.system(
            f"bash {theta_src_path}/templates/theta_{app_type}_init.sh {dataset_name}"
        )


def build_deepcode(args, theta_src=False):
    """
    构建完整可复现的模型训练代码Docker容器环境。
    """

    # -------- Check Dockerfile.deepcode --------
    if not os.path.exists("Dockerfile.deepcode"):
        logger.error(f"Dockerfile.deepcode does not exists.")
        return
    training_args_file = "outputs/latest/training_args.json"
    if not os.path.exists(training_args_file):
        os.system(f"python run_{args.dataset_name}.py --do_new")
        if not os.path.exists(training_args_file):
            logger.error(f"{training_args_file} does not exists.")
            return

    # -------- deepcode路径 --------
    deepcode_dir = "deepcode"
    if os.path.exists(deepcode_dir):
        shutil.rmtree(deepcode_dir)
    os.makedirs(deepcode_dir)
    os.system(f"cp Dockerfile.deepcode ./{deepcode_dir}/")

    # -------- Copy theta source --------
    if theta_src:
        # -------- 检查THETA_SRC环工--------
        #  if not "THETA_SRC" in os.environ:
        #      logger.error(
        #          f"THETA_SRC environment variable must be set to theta src path."
        #      )
        #      return
        #  theta_src_path = os.environ['THETA_SRC']
        #  if not os.path.exists(os.path.join(theta_src_path, "__init__.py")):
        #      logger.error(
        #          f"Ensure theta source must be located in THETA_SRC({theta_src_path})"
        #      )
        #      return
        theta_src_path = get_theta_src_path()

        if theta_src_path:
            os.system(
                f"rsync -ar --exclude examples --exclude __pycache__ --exclude *.pyc  $THETA_SRC ./{deepcode_dir}/"
            )

    # -------- Copy model src --------
    if os.path.exists("README.md"):
        os.system(f"cp README.md *.py {deepcode_dir}/")
    #  cmd = f"rsync -rLptgoDv --exclude rawdata data {deepcode_dir}/"
    cmd = f"rsync -rLptgoD data {deepcode_dir}/"
    logger.info(cmd)
    os.system(cmd)
    cmd = f"rsync -rLptgoD --exclude best outputs/latest {deepcode_dir}/outputs/"
    logger.info(cmd)
    os.makedirs(f"{deepcode_dir}/outputs/latest")
    os.system(cmd)
    cmd = f"cp *.py Makefile.* *.sh *.json requirements.txt {deepcode_dir}/"
    logger.info(cmd)
    os.system(cmd)
    home_dir = os.environ['HOME']
    if os.path.exists(f"{home_dir}/.pip/pip.conf"):
        os.system(f"cp {home_dir}/.pip/pip.conf  {deepcode_dir}/")

    with open(training_args_file, 'r') as rd:
        training_args = json.load(open(training_args_file, 'r'))
    dataset_name = training_args['dataset_name']
    local_id = training_args['local_id']

    # -------- 备份deepcode至training/local_id --------
    backup_dir = f"training/{local_id}"
    #  assert not os.path.exists(backup_dir), f"{backup_dir} exists."
    if os.path.exists(backup_dir):
        shutil.rmtree(backup_dir)
        logger.warning(f"Remove exists backup dir: {backup_dir}.")

    if not os.path.exists("training"):
        os.makedirs("training")

    shutil.copytree(f"{deepcode_dir}", backup_dir)
    logger.warning(f"Backup {deepcode_dir} to training/{local_id}")

    # -------- 构建Docker镜像 --------
    #  for k, v in training_args.items():
    #      if not k.startswith('_'):
    #          print(f"{k}: {v}")
    cmd = f"docker build -f Dockerfile.deepcode -t {dataset_name}:{local_id} {deepcode_dir}"
    logger.info(f"cmd: {cmd}")
    os.system(cmd)


def run_deepcode(args):
    gpus = args.gpus
    if args.dataset_name and args.local_id:
        dataset_name = args.dataset_name
        local_id = args.local_id[0]
        model_path = args.model_path
    else:
        training_args_file = "outputs/latest/training_args.json"
        if not os.path.exists(training_args_file):
            logger.error(f"{training_args_file} does not exists.")
            return

        with open(training_args_file, 'r') as rd:
            training_args = json.load(open(training_args_file, 'r'))

        for k, v in training_args.items():
            if not k.startswith('_'):
                print(f"{k}: {v}")
        dataset_name = training_args['dataset_name']
        if args.local_id:
            local_id = args.local_id[0]
        else:
            local_id = training_args['local_id']
        if args.model_path:
            model_path = args.model_path
        else:
            model_path = training_args['model_path']

    brat_collections_dir = os.path.abspath(args.brat_collections_dir)
    pwd = os.environ['PWD']
    cmd = f"docker run --gpus {gpus} -it --rm -v {pwd}/outputs/saved_models:/deepcode/outputs/saved_models -v {pwd}/submissions:/deepcode/submissions -v {model_path}:{model_path} -v {brat_collections_dir}:/deepcode/brat/data/collections {dataset_name}:{local_id} /bin/bash"
    logger.info(f"cmd: {cmd}")
    os.system(cmd)


def pull_deepcode(args):
    if not "THETA_REGISTRY_URL" in os.environ:
        logger.error(
            f"THETA_REGISTRY_URL environment variable must be set to docker registry."
        )
        return
    theta_registry_url = os.environ['THETA_REGISTRY_URL']

    local_id = args.local_id[0]
    dataset_name = args.dataset_name
    pwd = os.environ['PWD']
    deepcode_name = f"{dataset_name}:{local_id}"
    cmd = f"docker pull {theta_registry_url}/{deepcode_name} && docker tag {theta_registry_url}/{deepcode_name} {deepcode_name}"
    logger.info(f"cmd: {cmd}")
    os.system(cmd)


def exec_deepcode(args):

    gpus = args.gpus
    local_id = args.local_id[0]
    dataset_name = args.dataset_name
    model_path = args.model_path
    pwd = os.environ['PWD']
    cmd = f"docker run --gpus {gpus} -it --rm -v {pwd}:/deepcode/outputs/saved_models -v {pwd}:/deepcode/submissions -v {model_path}:{model_path} {dataset_name}:{local_id} /bin/bash"
    logger.info(f"cmd: {cmd}")
    os.system(cmd)


def push_deepcode(args, gpus="device=0"):
    if not "THETA_REGISTRY_URL" in os.environ:
        logger.error(
            f"THETA_REGISTRY_URL environment variable must be set to docker registry."
        )
        return
    theta_registry_url = os.environ['THETA_REGISTRY_URL']

    training_args_file = "outputs/latest/training_args.json"
    if not os.path.exists(training_args_file):
        logger.error(f"{training_args_file} does not exists.")
        return

    with open(training_args_file, 'r') as rd:
        training_args = json.load(open(training_args_file, 'r'))

    for k, v in training_args.items():
        if not k.startswith('_'):
            print(f"{k}: {v}")
    dataset_name = training_args['dataset_name']
    local_id = training_args['local_id']
    model_path = training_args['model_path']

    #  registry_url = "registry.cn-shanghai.aliyuncs.com/upside-down"
    remote_image = f"{theta_registry_url}/{dataset_name}:{local_id}"

    cmd = f"docker tag {dataset_name}:{local_id} {remote_image} && docker push {remote_image}"
    logger.info(f"cmd: {cmd}")
    os.system(cmd)


def version(args):
    from theta.__version__ import __version__
    print(f"Version: {__version__}")


def home(args):

    home_dir = os.path.dirname(__file__)
    home_dir = os.path.realpath(home_dir)
    print(home_dir)


def main(args):
    find_models(args)

    if args.version:
        version(args)
    elif args.home:
        home(args)
    elif args.init:
        init(args)
    elif args.list:
        list_models(args)
    elif args.diff:
        diff_models(args)
    elif args.show:
        show_model(args)
    elif args.new:
        new_model(args)
    elif args.use:
        use_model(args)
    elif args.export_train_data:
        export_train_data(args)
    elif args.export_test_data:
        export_test_data(args)
    elif args.export_submit_data:
        export_submit_data(args)
    elif args.import_brat_data:
        import_brat_data(args)
    elif args.import_poplar_data:
        import_poplar_data(args)
    elif args.json_to_brat:
        json_to_brat(args)
    elif args.run_brat:
        run_brat(args)
    elif args.brat_to_json:
        brat_to_json(args)
    elif args.diff_ner_datasets:
        diff_ner_datasets(args)
    elif args.merge_ner_datasets:
        merge_ner_datasets(args)
    elif args.mix_ner_datasets:
        mix_ner_datasets(args)
    elif args.build_deepcode:
        build_deepcode(args, theta_src=True)
    elif args.run_deepcode:
        run_deepcode(args)
    elif args.push_deepcode:
        push_deepcode(args)
    elif args.pull_deepcode:
        pull_deepcode(args)
    elif args.exec_deepcode:
        exec_deepcode(args)
    else:
        print("Usage: theta [list|diff]")


def get_args():
    import argparse
    parser = argparse.ArgumentParser()

    parser.add_argument("--version", action='store_true')
    parser.add_argument("--home", action='store_true')
    parser.add_argument("--init", action='store_true')
    parser.add_argument("--new", action='store_true')
    parser.add_argument("--use", action='store_true')
    parser.add_argument("--diff", action='store_true')
    parser.add_argument("--list", action='store_true')
    parser.add_argument("--show", action='store_true')
    parser.add_argument("--detail", action='store_true')
    parser.add_argument("--output_dir", default="./outputs")
    parser.add_argument("--submissions_dir", default="./submissions")
    parser.add_argument("--local_id", action='append')
    parser.add_argument("--gpus", default="device=0")

    parser.add_argument("--run_brat", action='store_true')

    parser.add_argument("--app_type",
                        default='ner',
                        choices=['ner', 'glue', 'spo'])
    parser.add_argument("--dataset_name", default=None)
    parser.add_argument("--brat_data_dir", default=None)
    parser.add_argument("--brat_collections_dir",
                        default="brat_data/collections")
    parser.add_argument("--brat_port", default=8001)
    parser.add_argument(
        "--model_path",
        default="/opt/share/pretrained/pytorch/bert-base-chinese")
    parser.add_argument("--format",
                        default='json',
                        choices=['json', 'lines', 'brat', 'poplar'])
    parser.add_argument("--export_train_data", action='store_true')
    parser.add_argument("--export_test_data", action='store_true')
    parser.add_argument("--export_submit_data", action='store_true')
    parser.add_argument("--import_brat_data", action='store_true')
    parser.add_argument("--import_poplar_data", action='store_true')

    parser.add_argument("--json_to_brat", action='store_true')
    parser.add_argument("--brat_to_json", action='store_true')
    parser.add_argument("--dataset_file", default=None)
    parser.add_argument("--output", default=None)

    parser.add_argument("--diff_ner_datasets",
                        nargs=2,
                        help='Check difference of 2 datasets')
    parser.add_argument("--max_pages",
                        type=int,
                        default=20,
                        help="Max pages of brat export.")

    parser.add_argument("--merge_ner_datasets",
                        nargs="+",
                        help='Merge NER datasets')
    parser.add_argument("--min_dups",
                        type=int,
                        default=2,
                        help="Min duplicates of merge tags.")

    parser.add_argument("--mix_ner_datasets",
                        nargs="+",
                        help='Mix NER datasets')

    parser.add_argument("--build_deepcode",
                        action="store_true",
                        help="Build deep code docker image.")
    parser.add_argument("--run_deepcode",
                        action="store_true",
                        help="Run deep code docker.")
    parser.add_argument("--push_deepcode",
                        action="store_true",
                        help="Push deep code docker.")
    parser.add_argument("--pull_deepcode",
                        action="store_true",
                        help="Pull deep code docker.")

    parser.add_argument("--exec_deepcode",
                        action="store_true",
                        help="Run deep code docker.")

    # -------- ner dataset file --------
    subparsers = parser.add_subparsers(help="Commands")
    ner_dataset_parser = subparsers.add_parser("ner_dataset",
                                               help="NerDataset")
    ner_dataset_parser.add_argument("--show",
                                    action="store_true",
                                    help="Show info of NerDataset file.")

    args = parser.parse_args()

    return args


if __name__ == '__main__':
    args = get_args()
    main(args)



from transformers import BertModel,BertPreTrainedModel
import torch.nn as nn
import torch
import torch.nn.functional as F
from torch.nn import BCELoss
import random
random.seed(42)
device = torch.device('cuda')
BertLayerNorm = torch.nn.LayerNorm
def merge_function(inputs):
    output = inputs[0]
    for i in range(1, len(inputs)):
        output += inputs[i]
    return output / len(inputs)
def batch_gather(data:torch.Tensor,index:torch.Tensor):
    index = index.unsqueeze(-1)
    index = index.expand(data.size()[0],index.size()[1],data.size()[2])
    return torch.gather(data,1,index)
def extrac_subject(output, subject_ids):
    """根据subject_ids从output中取出subject的向量表征
    """
    start = batch_gather(output, subject_ids[:, :1])
    end = batch_gather(output, subject_ids[:, 1:])
    so_res = merge_function([start,end])
    # subject = torch.cat([start, end], 2)
    # return start,end
    return so_res

def extrac_subject_1(output, subject_ids):
    """根据subject_ids从output中取出subject的向量表征
    """
    start = batch_gather(output, subject_ids[:, :1])
    end = batch_gather(output, subject_ids[:, 1:])
    # so_res = merge_function([start,end])
    # subject = torch.cat([start, end], 2)
    return start,end
    # return so_res

class REModel_sbuject(BertPreTrainedModel):
    def __init__(self, config):
        super().__init__(config)
        self.num_labels = config.num_labels
        self.obj_labels = 110
        self.bert = BertModel(config)
        self.dropout = nn.Dropout(config.attention_probs_dropout_prob)
        self.LayerNorm = BertLayerNorm(config.hidden_size, eps=1e-12)
        self.classifier = nn.Linear(config.hidden_size,config.num_labels)
        self.obj_classifier = nn.Linear(config.hidden_size, self.obj_labels)
        self.sub_pos_emb = nn.Embedding(256, 768)
        self.relu = nn.ReLU()
        self.linear = nn.Linear(768, 768)
        self.init_weights()
    def forward(
        self,
        input_ids=None,
        attention_mask=None,
        token_type_ids=None,
        position_ids=None,
        head_mask=None,
        inputs_embeds=None,
        labels=None,
        subject_ids = None,
        batch_size = None,
        obj_labels = None,
        sub_train = False,
        obj_train = False
    ):

        outputs_1 = self.bert(
            input_ids=input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
        )
        sequence_output = outputs_1[0]
        sequence_output = self.dropout(sequence_output)
        if sub_train == True:
            logits = self.classifier(sequence_output)
            outputs = (logits,)   # add hidden states and attention if they are here
            loss_fct = BCELoss(reduction='none')
            # loss_fct = BCEWithLogitsLoss(reduction='none')
            loss_sig = nn.Sigmoid()
            # Only keep active parts of the loss
            active_logits = logits.view(-1, self.num_labels)
            active_logits = loss_sig(active_logits)
            active_logits = active_logits ** 2
            if labels is not None :
                active_labels = labels.view(-1, self.num_labels).float()
                loss = loss_fct(active_logits, active_labels)
                # loss = loss.view(-1,sequence_output.size()[1],2)
                loss = loss.view(batch_size, -1, 2)
                loss = torch.mean(loss, 2)
                loss = torch.sum(attention_mask * loss) / torch.sum(attention_mask)
                outputs = (loss,) + outputs
            else:
                outputs = active_logits
        if obj_train == True:
            hidden_states = outputs_1[2][-2]
            # hidden_states = self.dropout(hidden_states)
            loss_obj = BCELoss(reduction='none')
            loss_sig = nn.Sigmoid()
            # sub_pos_start = self.sub_pos_emb(subject_ids[:, :1]).to(device)
            # sub_pos_end = self.sub_pos_emb(subject_ids[:, 1:]).to(device)
            # subject_start,subject_end = extrac_subject_1(hidden_states, subject_ids)
            # subject = extrac_subject(hidden_states, subject_ids)
            # subject_start = subject_start.to(device)
            # subject_end = subject_end.to(device)
            # subject = (sub_pos_start + subject_start + sub_pos_end + subject_end).to(device)
            subject = extrac_subject(hidden_states, subject_ids).to(device)
            batch_token_ids_obj = torch.add(hidden_states, subject)
            # batch_token_ids_obj = self.LayerNorm(batch_token_ids_obj)
            # batch_token_ids_obj = self.dropout(batch_token_ids_obj)
            # batch_token_ids_obj = F.dropout(batch_token_ids_obj,p=0.5)
            # batch_token_ids_obj = self.relu(self.linear(batch_token_ids_obj))
            # batch_token_ids_obj = self.dropout(batch_token_ids_obj)
            obj_logits = self.obj_classifier(batch_token_ids_obj)
            # obj_logits = self.dropout(obj_logits)
            # obj_logits = F.dropout(obj_logits,p=0.4)
            obj_logits = loss_sig(obj_logits)
            obj_logits = obj_logits ** 4
            obj_outputs = (obj_logits,)
            if obj_labels is not None:
                # obj_loss = loss_obj(obj_logits.view(-1, hidden_states.size()[1], self.obj_labels // 2, 2), obj_labels.float())
                obj_loss = loss_obj(obj_logits.view(batch_size, -1, self.obj_labels // 2, 2), obj_labels.float())
                obj_loss = torch.sum(torch.mean(obj_loss, 3), 2)
                obj_loss = torch.sum(obj_loss * attention_mask) / torch.sum(attention_mask)
                s_o_loss = torch.add(obj_loss, loss)
                outputs_obj = (s_o_loss,) + obj_outputs
            else:
                # outputs_obj = obj_logits.view(-1,hidden_states.size()[1],self.obj_labels // 2 ,2)
                outputs_obj = obj_logits.view(batch_size, -1, self.obj_labels // 2, 2)
        if obj_train == True:
            return outputs ,outputs_obj # (loss), scores, (hidden_states), (attentions)
        else:
            return outputs
class REModel_sbuject_1(BertPreTrainedModel):
    def __init__(self, config):
        super().__init__(config)
        self.num_labels = config.num_labels
        self.obj_labels = 110
        self.bert = BertModel(config)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)
        self.LayerNorm = BertLayerNorm(config.hidden_size, eps=1e-12)
        self.classifier = nn.Linear(config.hidden_size,config.num_labels)
        self.obj_classifier = nn.Linear(config.hidden_size, self.obj_labels)
        self.sub_pos_emb = nn.Embedding(256, 768)
        self.init_weights()
    def forward(
        self,
        input_ids=None,
        attention_mask=None,
        token_type_ids=None,
        position_ids=None,
        head_mask=None,
        inputs_embeds=None,
        labels=None,
        subject_ids = None,
        batch_size = None,
        obj_labels = None,
        sub_train = False,
        obj_train = False
    ):

        outputs_1 = self.bert(
            input_ids=input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
        )
        sequence_output = outputs_1[0]
        sequence_output = self.dropout(sequence_output)
        if sub_train == True:
            logits = self.classifier(sequence_output)
            outputs = (logits,)   # add hidden states and attention if they are here
            loss_fct = BCELoss(reduction='none')
            # loss_fct = BCEWithLogitsLoss(reduction='none')
            loss_sig = nn.Sigmoid()
            # Only keep active parts of the loss
            active_logits = logits.view(-1, self.num_labels)
            active_logits = loss_sig(active_logits)
            active_logits = active_logits ** 2
            if labels is not None :
                active_labels = labels.view(-1, self.num_labels).float()
                loss = loss_fct(active_logits, active_labels)
                loss = loss.view(batch_size,-1,2)
                loss = torch.mean(loss, 2)
                loss = torch.sum(attention_mask * loss) / torch.sum(attention_mask)
                outputs = (loss,) + outputs
            else:
                outputs = active_logits
        if obj_train == True:
            hidden_states = outputs_1[2][-2]
            loss_obj = BCELoss(reduction='none')
            sub_pos_start = self.sub_pos_emb(subject_ids[:, :1])
            sub_pos_end = self.sub_pos_emb(subject_ids[:, 1:])
            sub_pos = sub_pos_start + sub_pos_end
            # sub_pos = sub_pos.expand(batch_size,)
            loss_sig = nn.Sigmoid()
            subject = extrac_subject(hidden_states, subject_ids)
            # subject = subject_start + subject_end
            # subject = torch.add(subject,sub_pos)
            batch_token_ids_obj = torch.add(hidden_states, subject)
            batch_token_ids_obj = self.LayerNorm(batch_token_ids_obj)
            batch_token_ids_obj = self.dropout(batch_token_ids_obj)
            obj_logits = self.obj_classifier(batch_token_ids_obj)
            obj_logits = F.dropout(obj_logits,p=0.4)
            obj_logits = loss_sig(obj_logits)
            obj_logits = obj_logits ** 4
            obj_outputs = (obj_logits,)
            if obj_labels is not None:
                obj_loss = loss_obj(obj_logits.view(batch_size, -1, self.obj_labels // 2, 2), obj_labels.float())
                obj_loss = torch.sum(torch.mean(obj_loss, 3), 2)
                obj_loss = torch.sum(obj_loss * attention_mask) / torch.sum(attention_mask)
                s_o_loss = torch.add(obj_loss, loss)
                outputs_obj = (s_o_loss,) + obj_outputs
            else:
                outputs_obj = obj_logits.view(batch_size,-1,self.obj_labels // 2 ,2)
        if obj_train == True:
            return outputs ,outputs_obj # (loss), scores, (hidden_states), (attentions)
        else:
            return outputs
class REModel_sbuject_2(BertPreTrainedModel):
    def __init__(self, config):
        super().__init__(config)
        self.num_labels = config.num_labels
        self.obj_labels = 110
        self.bert = BertModel(config)
        self.linear = nn.Linear(768, 768)
        self.dropout = nn.Dropout(config.attention_probs_dropout_prob)
        self.LayerNorm = BertLayerNorm(config.hidden_size, eps=1e-12)
        self.classifier = nn.Linear(config.hidden_size,config.num_labels)
        self.obj_classifier = nn.Linear(config.hidden_size, self.obj_labels)
        self.sub_pos_emb = nn.Embedding(256, 768)
        self.relu = nn.ReLU()
        self.init_weights()
    def forward(
        self,
        input_ids=None,
        attention_mask=None,
        token_type_ids=None,
        position_ids=None,
        head_mask=None,
        inputs_embeds=None,
        labels=None,
        subject_ids = None,
        batch_size = None,
        obj_labels = None,
        sub_train = False,
        obj_train = False
    ):

        outputs_1 = self.bert(
            input_ids=input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
        )
        sequence_output = outputs_1[0]
        sequence_output = self.dropout(sequence_output)
        if sub_train == True:
            logits = self.classifier(sequence_output)
            outputs = (logits,)   # add hidden states and attention if they are here
            loss_fct = BCELoss(reduction='none')
            # loss_fct = BCEWithLogitsLoss(reduction='none')
            loss_sig = nn.Sigmoid()
            # Only keep active parts of the loss
            active_logits = logits.view(-1, self.num_labels)
            active_logits = loss_sig(active_logits)
            active_logits = active_logits ** 2
            if labels is not None :
                active_labels = labels.view(-1, self.num_labels).float()
                loss = loss_fct(active_logits, active_labels)
                # loss = loss.view(-1,sequence_output.size()[1],2)
                loss = loss.view(batch_size, -1, 2)
                loss = torch.mean(loss, 2)
                loss = torch.sum(attention_mask * loss) / torch.sum(attention_mask)
                outputs = (loss,) + outputs
            else:
                outputs = active_logits
        if obj_train == True:
            hidden_states = outputs_1[2][-2]
            hidden_states_1 = outputs_1[2][-3]
            # hidden_states = self.dropout(hidden_states)
            loss_obj = BCELoss(reduction='none')
            loss_sig = nn.Sigmoid()
            sub_pos_start = self.sub_pos_emb(subject_ids[:, :1]).to(device)
            sub_pos_end = self.sub_pos_emb(subject_ids[:, 1:]).to(device)
            subject_start_last, subject_end_last = extrac_subject_1(sequence_output, subject_ids)
            subject_start_1,subject_end_1 = extrac_subject_1(hidden_states_1, subject_ids)
            subject_start,subject_end = extrac_subject_1(hidden_states, subject_ids)
            # subject = extrac_subject(hidden_states, subject_ids)
            # subject_start = subject_start.to(device)
            # subject_end = subject_end.to(device)
            subject = (sub_pos_start + subject_start + sub_pos_end + subject_end + subject_start_last + subject_start_1 + subject_end_1 + subject_end_1).to(device)
            # subject = extrac_subject(sequence_output, subject_ids).to(device)
            batch_token_ids_obj = torch.add(hidden_states, subject)
            batch_token_ids_obj = self.LayerNorm(batch_token_ids_obj)
            batch_token_ids_obj = self.dropout(batch_token_ids_obj)
            # batch_token_ids_obj = F.dropout(batch_token_ids_obj,p=0.5)
            batch_token_ids_obj = self.relu(self.linear(batch_token_ids_obj))
            batch_token_ids_obj = self.dropout(batch_token_ids_obj)
            # batch_token_ids_obj = F.dropout(batch_token_ids_obj,p=0.4)
            obj_logits = self.obj_classifier(batch_token_ids_obj)

            obj_logits = loss_sig(obj_logits)
            obj_logits = obj_logits ** 4
            obj_outputs = (obj_logits,)
            if obj_labels is not None:
                # obj_loss = loss_obj(obj_logits.view(-1, hidden_states.size()[1], self.obj_labels // 2, 2), obj_labels.float())
                obj_loss = loss_obj(obj_logits.view(batch_size, -1, self.obj_labels // 2, 2), obj_labels.float())
                obj_loss = torch.sum(torch.mean(obj_loss, 3), 2)
                obj_loss = torch.sum(obj_loss * attention_mask) / torch.sum(attention_mask)
                s_o_loss = torch.add(obj_loss, loss)
                outputs_obj = (s_o_loss,) + obj_outputs
            else:
                # outputs_obj = obj_logits.view(-1,hidden_states.size()[1],self.obj_labels // 2 ,2)
                outputs_obj = obj_logits.view(batch_size, -1, self.obj_labels // 2, 2)
        if obj_train == True:
            return outputs ,outputs_obj # (loss), scores, (hidden_states), (attentions)
        else:
            return outputs


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
from tqdm import tqdm
from loguru import logger
import torch


def batch_encode_texts(all_texts, label2id, tokenizer, max_seq_length):
    all_encodes = tokenizer.batch_encode(all_texts, add_special_tokens=True)
    all_tokens = all_encodes['tokens']
    all_token_offsets = all_encodes['offsets']
    all_token2char = all_encodes['token2char']
    all_char2token = all_encodes['char2token']

    all_input_ids = all_encodes['ids']
    all_attention_mask = all_encodes['attention_mask']
    all_token_type_ids = all_encodes['type_ids']

    all_input_lens = [len(tokens) for tokens in all_tokens]

    all_padding_lens = [max_seq_length - n for n in all_input_lens]
    for i, (input_ids, attention_mask, token_type_ids, token2char,
            token_offsets, padding_length) in enumerate(
                tqdm(zip(all_input_ids, all_attention_mask, all_token_type_ids,
                         all_token2char, all_token_offsets, all_padding_lens),
                     desc="common_batch_encode")):

        if padding_length < 0:
            all_tokens[i] = all_tokens[i][:max_seq_length - 1] + ['[SEP]']
            all_input_ids[i] = input_ids[:max_seq_length - 1] + [102]
            all_attention_mask[i] = attention_mask[:max_seq_length]
            all_token_type_ids[i] = token_type_ids[:max_seq_length]
            all_token2char[i] = token2char[:max_seq_length]
            all_token_offsets[i] = token_offsets[:max_seq_length - 1] + [(0, 0)
                                                                         ]
            #  logger.warning(f"all_tokens[i]: {all_tokens[i]}")
            #  logger.debug(f"all_input_ids[i]: {all_input_ids[i]}")
            #  logger.debug(f"all_token_offsets[i]: {all_token_offsets[i]}")

        else:
            all_input_ids[i] = input_ids + [0] * padding_length
            all_attention_mask[i] = attention_mask + [0] * padding_length
            all_token_type_ids[i] = token_type_ids + [0] * padding_length
            all_token2char[i] = token2char + [0] * padding_length
            all_token_offsets[i] = token_offsets + [(0, 0)] * padding_length
        all_input_lens[i] = len(input_ids)

        #  logger.debug(
        #      f"padding_length: {padding_length}, len(input_ids): {len(input_ids)}"
        #  )
        assert len(all_input_ids[i]) == max_seq_length
        assert len(all_attention_mask[i]) == max_seq_length
        assert len(all_token_type_ids[i]) == max_seq_length
        assert len(all_token2char[i]) == max_seq_length
        assert len(all_token_offsets[i]) == max_seq_length

    #  logger.warning(f"all_input_ids: {all_input_ids}")
    all_input_ids = torch.from_numpy(np.array(all_input_ids, dtype=np.int64))
    all_attention_mask = torch.from_numpy(
        np.array(all_attention_mask, dtype=np.int64))
    all_token_type_ids = torch.from_numpy(
        np.array(all_token_type_ids, dtype=np.int64))
    all_input_lens = torch.from_numpy(np.array(all_input_lens, dtype=np.int64))
    all_token_offsets = torch.from_numpy(
        np.array(all_token_offsets, dtype=np.int64))

    return (all_tokens, all_token2char, all_char2token, all_input_ids,
            all_attention_mask, all_token_type_ids, all_input_lens,
            all_token_offsets)


def batch_encode_tags(all_tags, all_token_lens, all_char2token, label2id,
                      max_seq_length):
    def encode_tags(tags, num_tokens, char2token):
        #  logger.warning(f"num_tokens: {num_tokens}")
        start_ids = [label2id['[unused1]']] * num_tokens
        end_ids = [label2id['[unused1]']] * num_tokens

        encoded_tags = []
        for tag in tags:
            label = tag.category
            start = tag.start
            end = tag.start + len(tag.mention)

            if start < 0 or start >= len(char2token):
                continue
            if end < 0 or end >= len(char2token):
                continue
            start = char2token[start]
            end = char2token[end]

            if start >= num_tokens:
                continue
            if end >= num_tokens:
                continue

            start_ids[start] = label2id[label]
            end_ids[end] = label2id[label]
            encoded_tags.append((label2id[label], start, end))

        return encoded_tags, start_ids, end_ids

    all_encoded_tags = []
    all_start_ids = []
    all_end_ids = []

    for tags, num_tokens, char2token in zip(all_tags, all_token_lens,
                                            all_char2token):

        encoded_tags, start_ids, end_ids = encode_tags(tags, num_tokens,
                                                       char2token)

        assert len(start_ids) == max_seq_length
        assert len(end_ids) == max_seq_length

        all_encoded_tags.append(encoded_tags)
        all_start_ids.append(start_ids)
        all_end_ids.append(end_ids)

    return all_start_ids, all_end_ids, all_encoded_tags


def common_to_tensors(all_input_ids, all_attention_mask, all_token_type_ids,
                      all_input_lens, all_token_offsets):
    all_input_ids = torch.from_numpy(np.array(all_input_ids, dtype=np.int64))
    all_attention_mask = torch.from_numpy(
        np.array(all_attention_mask, dtype=np.int64))
    all_token_type_ids = torch.from_numpy(
        np.array(all_token_type_ids, dtype=np.int64))
    all_input_lens = torch.from_numpy(np.array(all_input_lens, dtype=np.int64))
    all_token_offsets = torch.from_numpy(
        np.array(all_token_offsets, dtype=np.int64))

    return all_input_ids, all_attention_mask, all_token_type_ids, all_input_lens, all_token_offsets



# -*- coding: utf-8 -*-
# File: svhn.py


import numpy as np
import os

from ...utils import logger
from ...utils.fs import download, get_dataset_path
from ..base import RNGDataFlow

__all__ = ['SVHNDigit']

SVHN_URL = "http://ufldl.stanford.edu/housenumbers/"


class SVHNDigit(RNGDataFlow):
    """
    `SVHN <http://ufldl.stanford.edu/housenumbers/>`_ Cropped Digit Dataset.
    Produces [img, label], img of 32x32x3 in range [0,255], label of 0-9
    """
    _Cache = {}

    def __init__(self, name, data_dir=None, shuffle=True):
        """
        Args:
            name (str): 'train', 'test', or 'extra'.
            data_dir (str): a directory containing the original {train,test,extra}_32x32.mat.
            shuffle (bool): shuffle the dataset.
        """
        self.shuffle = shuffle

        if name in SVHNDigit._Cache:
            self.X, self.Y = SVHNDigit._Cache[name]
            return
        if data_dir is None:
            data_dir = get_dataset_path('svhn_data')
        assert name in ['train', 'test', 'extra'], name
        filename = os.path.join(data_dir, name + '_32x32.mat')
        if not os.path.isfile(filename):
            url = SVHN_URL + os.path.basename(filename)
            logger.info("File {} not found!".format(filename))
            logger.info("Downloading from {} ...".format(url))
            download(url, os.path.dirname(filename))
        logger.info("Loading {} ...".format(filename))
        data = scipy.io.loadmat(filename)
        self.X = data['X'].transpose(3, 0, 1, 2)
        self.Y = data['y'].reshape((-1))
        self.Y[self.Y == 10] = 0
        SVHNDigit._Cache[name] = (self.X, self.Y)

    def __len__(self):
        return self.X.shape[0]

    def __iter__(self):
        n = self.X.shape[0]
        idxs = np.arange(n)
        if self.shuffle:
            self.rng.shuffle(idxs)
        for k in idxs:
            # since svhn is quite small, just do it for safety
            yield [self.X[k], self.Y[k]]

    @staticmethod
    def get_per_pixel_mean(names=('train', 'test', 'extra')):
        """
        Args:
            names (tuple[str]): names of the dataset split

        Returns:
            a 32x32x3 image, the mean of all images in the given datasets
        """
        for name in names:
            assert name in ['train', 'test', 'extra'], name
        images = [SVHNDigit(x).X for x in names]
        return np.concatenate(tuple(images)).mean(axis=0)


try:
    import scipy.io
except ImportError:
    from ...utils.develop import create_dummy_class
    SVHNDigit = create_dummy_class('SVHNDigit', 'scipy.io')  # noqa

if __name__ == '__main__':
    a = SVHNDigit('train')
    b = SVHNDigit.get_per_pixel_mean()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import copy
import json
import random
from dataclasses import dataclass, field

from loguru import logger
from tqdm import tqdm

from theta.utils import seg_generator

from ..utils import seg_generator

#  import mlflow


class InputExample(object):
    """A single training/test example for token classification."""
    def __init__(self, guid, text, spo_list, text_offset=0):
        self.guid = guid
        self.text = text
        self.spo_list = spo_list
        self.text_offset = text_offset

    def __repr__(self):
        return str(self.to_json_string())

    def to_dict(self):
        """Serializes this instance to a Python dictionary."""
        output = copy.deepcopy(self.__dict__)
        return output

    def to_json_string(self):
        """Serializes this instance to a JSON string."""
        return json.dumps(self.to_dict(), indent=2, sort_keys=True) + "\n"


def load_spo_examples(data_generator, examples_file, seg_len=0, seg_backoff=0):

    examples = []

    for guid, text, text_b, spo_list in data_generator(examples_file):
        assert text is not None
        for (seg_text, ), text_offset in seg_generator((text, ), seg_len,
                                                       seg_backoff):

            examples.append(
                InputExample(guid=guid,
                             text=seg_text,
                             spo_list=spo_list,
                             text_offset=text_offset))
    logger.info(f"Loaded {len(examples)} examples.")

    return examples


def show_spo_datainfo(predicate_labels, train_data_generator, train_file,
                      test_data_generator, test_file):
    train_lengths = [
        len(text) for guid, text, _, label in train_data_generator(train_file)
    ]

    test_lengths = [
        len(text) for guid, text, _, _ in test_data_generator(test_file)
    ]

    logger.info(f"****** predicate_labels ******")
    logger.info(f"{predicate_labels}")
    logger.info(f"")
    import numpy as np
    logger.info(f"****** train lengths ******")
    logger.info(f"mean: {np.mean(train_lengths):.2f}")
    logger.info(f"std: {np.std(train_lengths):.2f}")
    logger.info(f"min: {np.min(train_lengths)}")
    logger.info(f"max: {np.max(train_lengths)}")
    logger.info(f"")

    logger.info(f"****** test lengths ******")
    logger.info(f"mean: {np.mean(test_lengths):.2f}")
    logger.info(f"std: {np.std(test_lengths):.2f}")
    logger.info(f"min: {np.min(test_lengths)}")
    logger.info(f"max: {np.max(test_lengths)}")
    logger.info(f"")


def get_spo_preds_reviews(preds, examples, seg_len, seg_backoff):
    reviews = {}
    category_mentions = {}
    for spoes, input_example in tqdm(zip(preds, examples)):
        guid = input_example.guid
        text_offset = input_example.text_offset
        text = input_example.text

        spo_list = []
        for x in spoes:
            (sub_start, sub_mention), predicate, (obj_start, obj_mention) = x
            spo = {
                'subject': (sub_start + text_offset, sub_mention),
                'predicate': predicate,
                'object': (obj_start + text_offset, obj_mention)
            }
            spo_list.append(spo)

        #  logger.info(
        #      f"guid: {guid}, offset: {text_offset}, entities: {entities}")

        if guid not in reviews:
            reviews[guid] = {
                'guid': guid,
                'text': "",
                'annotated_text': "",
                'spo_list': []
            }
        reviews[guid]['text'] += text[:seg_len - seg_backoff]
        #  reviews[guid]['annotated_text'] += annotated_text
        reviews_spo_list = reviews[guid]['spo_list']
        reviews_spo_list.extend(spo_list)

        reviews_spo_list = sorted(
            reviews_spo_list,
            key=lambda x: x['subject'][0] * 10000 + x['object'][0])
        from ..utils import remove_duplicate_entities
        reviews_spo_list = remove_duplicate_entities(reviews_spo_list)

        reviews[guid]['spo_list'] = reviews_spo_list
        #  logger.info(f"guid: {guid}, tags: {reviews[guid]['tags']}")

    return reviews, category_mentions


def save_spo_preds(args, preds, test_examples):
    assert len(test_examples) == len(preds)

    reviews, category_mentions = get_spo_preds_reviews(preds, test_examples,
                                                       args.seg_len,
                                                       args.seg_backoff)
    reviews_file = f"{args.latest_dir}/{args.dataset_name}_reviews_{args.local_id}.json"
    json.dump(reviews, open(reviews_file, 'w'), ensure_ascii=False, indent=2)
    logger.info(f"Reviews file: {reviews_file}, {len(reviews)} examples.")

    logger.info(f"Total {len(preds)} lines saved to {reviews_file}")

    #  ----- Tracking -----
    #  if args.do_experiment:
    #      logger.debug(f"mlflow tracking uri: {mlflow.get_tracking_uri()}")
    #      mlflow.log_param(f"{args.dataset_name}_reviews_file", reviews_file)
    #      mlflow.log_artifact(reviews_file, args.artifact_path)

    return reviews_file


# 输入：train_data_generator -> (guid, text, None, tags)
#       tag: ((sub_start, sub_mention), predicate, (obj_start, obj_mention))
#       predicate: <sub_type>_<predicate_type>_<obj_type>
# 任务：
#       1. text长文本切分(seg_len, seg_backoff)
#       2. 数据增强
#       3. 按行划分为训练集和验证集


def fix_mention_with_blank(start, mention):
    real_mention = mention.strip()
    if real_mention:
        if real_mention == mention:
            return start, mention
        p0 = mention.find(real_mention)
        assert p0 >= 0, f"mention: {mention}, real_mention: {real_mention}"
        logger.warning(
            f"Fix |{start}|{mention}| to |{start+p0}:{real_mention}|")
        start += p0
        mention = real_mention
        return start, mention
    else:
        return -1, None


def load_spo_labeled_examples(data_generator,
                              predicate_labels,
                              seg_len=0,
                              seg_backoff=0,
                              num_augments=0,
                              allow_overlap=False):

    examples = []
    for guid, text, _, tags in data_generator(None):

        for (seg_text, ), seg_offset in seg_generator((text, ), seg_len,
                                                      seg_backoff):
            # 按text_offset过滤tags

            seg_start = seg_offset
            seg_end = seg_offset + min(seg_len, len(seg_text))
            #  logger.info(f"tags: {tags}")
            #  seg_tags = [
            #      ((x[0][0] - seg_start, x[0][1]), x[1], (x[2][0] - seg_start,
            #                                              x[2][1])) for x in tags
            #      if x[0][0] >= seg_start and x[0][0] + len(x[0][1]) < seg_end
            #      and x[2][0] >= seg_start and x[2][0] + len(x[2][1]) < seg_end
            #  ]
            seg_tags = []
            for x in tags:
                #  s_start, s_mention = x[0]
                #  predicate = x[1]
                #  o_start, o_mention = x[2]

                s_category = x['sub']['category']
                s_start = x['sub']['start']
                s_mention = x['sub']['mention']
                predicate = x['predicate']
                o_category = x['obj']['category']
                o_start = x['obj']['start']
                o_mention = x['obj']['mention']

                #  logger.warning(
                #      f"({s_category}, {s_start}, {s_mention}), {predicate}, ({o_category}, {o_start}, {o_mention})"
                #  )
                #  for seg_tag in seg_tags:
                #  (s_start, s_mention), predicate, (o_start, o_mention) = seg_tag

                s_start, s_mention = fix_mention_with_blank(s_start, s_mention)
                if s_start < 0:
                    continue
                o_start, o_mention = fix_mention_with_blank(o_start, o_mention)
                if o_start < 0:
                    continue
                s_end = s_start + len(s_mention)
                o_end = o_start + len(o_mention)

                if s_start >= seg_start and s_end < seg_end and o_start >= seg_start and o_end < seg_end:
                    s_start -= seg_start
                    o_start -= seg_start
                    s_end = s_start + len(s_mention)
                    o_end = o_start + len(o_mention)

                    assert seg_text[
                        s_start:
                        s_end] == s_mention, f"subject tag: |{seg_text[s_start:s_end]}| != mention: |{s_mention}|. seg: ({seg_start},{seg_end}), seg_tag: {((s_start, s_mention), predicate, (o_start, o_mention))}, seg_text: {seg_text}"
                    assert seg_text[
                        o_start:
                        o_end] == o_mention, f"object tag: |{seg_text[o_start:o_end]}| != mention: |{o_mention}|. seg: (){seg_start}:{seg_end}), seg_tag: {((s_start, s_mention), predicate, (o_start, o_mention))}, seg_text: {seg_text}"
                    seg_tags.append(((s_start, s_mention), predicate,
                                     (o_start, o_mention)))
            if seg_tags:
                examples.append(
                    InputExample(guid=guid,
                                 text=seg_text,
                                 spo_list=seg_tags,
                                 text_offset=seg_start))

    logger.info(f"Loaded {len(examples)} examples.")
    return examples


def load_train_val_examples(args,
                            train_data_generator,
                            predicate_labels,
                            shuffle=True,
                            train_rate=0.9,
                            num_augments=0):
    lines = []

    train_base_examples = load_spo_labeled_examples(
        train_data_generator,
        predicate_labels,
        seg_len=args.seg_len,
        seg_backoff=args.seg_backoff,
        num_augments=num_augments,
        allow_overlap=args.allow_overlap)

    if args.train_sample_rate < 1.0:
        num_samples = int(len(train_base_examples) * args.train_sample_rate)
        train_base_examples = train_base_examples[:num_samples]

    from ..utils import split_train_eval_examples
    train_examples, val_examples = split_train_eval_examples(
        train_base_examples,
        train_rate=train_rate,
        fold=args.fold,
        shuffle=shuffle)

    logger.info(f"Loaded {len(train_examples)} train examples, "
                f"{len(val_examples)} val examples.")
    return train_examples, val_examples


def load_test_examples(args, test_data_generator):
    from theta.modeling import load_spo_examples
    test_base_examples = load_spo_examples(test_data_generator,
                                           args.test_file,
                                           seg_len=args.seg_len,
                                           seg_backoff=args.seg_backoff)

    logger.info(f"Loaded {len(test_base_examples)} test examples.")
    return test_base_examples



# -*- coding: utf-8 -*-
# File: loadcaffe.py


import numpy as np
import os
import sys

from . import logger
from .concurrency import subproc_call
from .fs import download, get_dataset_path
from .utils import change_env

__all__ = ['load_caffe', 'get_caffe_pb']

CAFFE_PROTO_URL = "https://github.com/BVLC/caffe/raw/master/src/caffe/proto/caffe.proto"


class CaffeLayerProcessor(object):

    def __init__(self, net):
        self.net = net
        self.layer_names = net._layer_names
        self.param_dict = {}
        self.processors = {
            'Convolution': self.proc_conv,
            'InnerProduct': self.proc_fc,
            'BatchNorm': self.proc_bn,
            'Scale': self.proc_scale
        }

    def process(self):
        for idx, layer in enumerate(self.net.layers):
            param = layer.blobs
            name = self.layer_names[idx]
            if layer.type in self.processors:
                logger.info("Processing layer {} of type {}".format(
                    name, layer.type))
                dic = self.processors[layer.type](idx, name, param)
                self.param_dict.update(dic)
            elif len(layer.blobs) != 0:
                logger.warn(
                    "{} layer contains parameters but is not supported!".format(layer.type))
        return self.param_dict

    def proc_conv(self, idx, name, param):
        assert len(param) <= 2
        assert param[0].data.ndim == 4
        # caffe: ch_out, ch_in, h, w
        W = param[0].data.transpose(2, 3, 1, 0)
        if len(param) == 1:
            return {name + '/W': W}
        else:
            return {name + '/W': W,
                    name + '/b': param[1].data}

    def proc_fc(self, idx, name, param):
        # TODO caffe has an 'transpose' option for fc/W
        assert len(param) == 2
        prev_layer_name = self.net.bottom_names[name][0]
        prev_layer_output = self.net.blobs[prev_layer_name].data
        if prev_layer_output.ndim == 4:
            logger.info("FC layer {} takes spatial data.".format(name))
            W = param[0].data
            # original: outx(CxHxW)
            W = W.reshape((-1,) + prev_layer_output.shape[1:]).transpose(2, 3, 1, 0)
            # become: (HxWxC)xout
        else:
            W = param[0].data.transpose()
        return {name + '/W': W,
                name + '/b': param[1].data}

    def proc_bn(self, idx, name, param):
        scale_factor = param[2].data[0]
        return {name + '/mean/EMA': param[0].data / scale_factor,
                name + '/variance/EMA': param[1].data / scale_factor}

    def proc_scale(self, idx, name, param):
        bottom_name = self.net.bottom_names[name][0]
        # find the bn layer before this scaling
        for i, layer in enumerate(self.net.layers):
            if layer.type == 'BatchNorm':
                name2 = self.layer_names[i]
                bottom_name2 = self.net.bottom_names[name2][0]
                if bottom_name2 == bottom_name:
                    # scaling and BN share the same bottom, should merge
                    logger.info("Merge {} and {} into one BatchNorm layer".format(
                        name, name2))
                    return {name2 + '/beta': param[1].data,
                            name2 + '/gamma': param[0].data}
        # assume this scaling layer is part of some BN
        logger.error("Could not find a BN layer corresponding to this Scale layer!")
        raise ValueError()


def load_caffe(model_desc, model_file):
    """
    Load a caffe model. You must be able to ``import caffe`` to use this
    function.

    Args:
        model_desc (str): path to caffe model description file (.prototxt).
        model_file (str): path to caffe model parameter file (.caffemodel).
    Returns:
        dict: the parameters.
    """
    with change_env('GLOG_minloglevel', '2'):
        import caffe
        caffe.set_mode_cpu()
        net = caffe.Net(model_desc, model_file, caffe.TEST)
    param_dict = CaffeLayerProcessor(net).process()
    logger.info("Model loaded from caffe. Params: " +
                ", ".join(sorted(param_dict.keys())))
    return param_dict


def get_caffe_pb():
    """
    Get caffe protobuf.

    Returns:
        The imported caffe protobuf module.
    """
    dir = get_dataset_path('caffe')
    caffe_pb_file = os.path.join(dir, 'caffe_pb2.py')
    if not os.path.isfile(caffe_pb_file):
        download(CAFFE_PROTO_URL, dir)
        assert os.path.isfile(os.path.join(dir, 'caffe.proto'))

        cmd = "protoc --version"
        version, ret = subproc_call(cmd, timeout=3)
        if ret != 0:
            sys.exit(1)
        try:
            version = version.decode('utf-8')
            version = float('.'.join(version.split(' ')[1].split('.')[:2]))
            assert version >= 2.7, "Require protoc>=2.7 for Python3"
        except Exception:
            logger.exception("protoc --version gives: " + str(version))
            raise

        cmd = 'cd {} && protoc caffe.proto --python_out .'.format(dir)
        ret = os.system(cmd)
        assert ret == 0, \
            "Command `{}` failed!".format(cmd)
        assert os.path.isfile(caffe_pb_file), caffe_pb_file
    import imp
    return imp.load_source('caffepb', caffe_pb_file)


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('model', help='.prototxt file')
    parser.add_argument('weights', help='.caffemodel file')
    parser.add_argument('output', help='output npz file')
    args = parser.parse_args()
    ret = load_caffe(args.model, args.weights)

    if args.output.endswith('.npz'):
        np.savez_compressed(args.output, **ret)
    elif args.output.endswith('.npy'):
        logger.warn("Please use npz format instead!")
        np.save(args.output, ret)
    else:
        raise ValueError("Unknown format {}".format(args.output))
